"""Energy Distance module."""

from typing import Optional, Union

import numpy as np  # type: ignore
from scipy.stats import energy_distance  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import UnivariateData
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBased,
    DistanceResult,
)


class EnergyDistance(BaseDistanceBased):
    """EnergyDistance [szekely2013energy]_ detector.

    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]
    :param kwargs: additional keyword arguments to pass to scipy.stats.energy_distance
    :type kwargs: Dict[str, Any]

    :References:

    .. [szekely2013energy] Székely, Gábor J., and Maria L. Rizzo.
        "Energy statistics: A class of statistics based on distances."
        Journal of statistical planning and inference 143.8 (2013): 1249-1272.

    :Example:

    >>> from frouros.detectors.data_drift import EnergyDistance
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = EnergyDistance()
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    DistanceResult(distance=0.8359206395514527)
    """  # noqa: E501

    def __init__(  # noqa: D107
        self,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            statistical_type=UnivariateData(),
            statistical_method=self._energy_distance,
            statistical_kwargs=kwargs,
            callbacks=callbacks,
        )
        self.kwargs = kwargs

    def _distance_measure(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        emd = self._energy_distance(X=X_ref, Y=X, **self.kwargs)
        distance = DistanceResult(distance=emd)
        return distance

    @staticmethod
    def _energy_distance(
        X: np.ndarray,  # noqa: N803
        Y: np.ndarray,
        **kwargs,
    ) -> float:
        energy = energy_distance(
            u_values=X.flatten(),
            v_values=Y.flatten(),
            **kwargs,
        )
        return energy



"""KSWIN (Kolmogorov-Smirnov Windowing) module."""

import itertools
from collections import deque
from typing import Optional, Union

import numpy as np  # type: ignore
from scipy.stats import ks_2samp  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.window_based.base import (
    BaseWindowConfig,
    BaseWindow,
)


class KSWINConfig(BaseWindowConfig):
    """KSWIN (Kolmogorov-Smirnov Windowing) [raab2020reactive]_ configuration.

    :param alpha: significance value, defaults to 0.0001
    :type alpha: float
    :param seed: seed value, defaults to None
    :type seed: Optional[int]
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 100
    :type min_num_instances: int
    :param num_test_instances: numbers of instances to be used by the statistical test, defaults to 30
    :type num_test_instances: int
    :raises ValueError: Value error exception if seed is not valid

    :References:

    .. [raab2020reactive] Raab, Christoph, Moritz Heusinger, and Frank-Michael Schleif.
        "Reactive soft prototype computing for concept drift streams."
        Neurocomputing 416 (2020): 340-351.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        alpha: float = 0.0001,
        seed: Optional[int] = None,
        min_num_instances: int = 100,
        num_test_instances: int = 30,
    ) -> None:
        try:
            np.random.seed(seed=seed)
        except ValueError as e:
            raise e
        super().__init__(min_num_instances=min_num_instances)
        self.alpha = alpha
        self.num_test_instances = num_test_instances

    @property
    def alpha(self) -> float:
        """Significance value property.

        :return: significance value
        :rtype: float
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value: int) -> None:
        """Significance value setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value <= 0:
            raise ValueError("alpha value must be greater than 0.")
        self._alpha = value

    @property
    def num_test_instances(self) -> int:
        """Number of tests instances property.

        :return: number of tests instances to be used by the statistical test
        :rtype: int
        """
        return self._num_test_instances

    @num_test_instances.setter
    def num_test_instances(self, value: int) -> None:
        """Number of tests instances value setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value > self.min_num_instances:
            raise ValueError(
                "num_test_instances value must be smaller or equal than "
                "min_num_instances."
            )
        if value < 1:
            raise ValueError("num_test_instances value must be greater than 0.")
        self._num_test_instances = value


class KSWIN(BaseWindow):
    """KSWIN (Kolmogorov-Smirnov Windowing) [raab2020reactive]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`KSWINConfig` is used.
    :type config: Optional[KSWINConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :References:

    .. [raab2020reactive] Raab, Christoph, Moritz Heusinger, and Frank-Michael Schleif.
        "Reactive soft prototype computing for concept drift streams."
        Neurocomputing 416 (2020): 340-351.

    :Example:

    >>> from frouros.detectors.concept_drift import KSWIN, KSWINConfig
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.normal(loc=0.2, scale=0.01, size=1000)
    >>> dist_b = np.random.normal(loc=0.8, scale=0.04, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = KSWIN(config=KSWINConfig(seed=31))
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    Change detected at step 1016
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = KSWINConfig

    def __init__(  # noqa: D107
        self,
        config: Optional[KSWINConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        self.additional_vars = {
            "window": deque(maxlen=self.config.min_num_instances),
        }
        self._set_additional_vars_callback()

    @property
    def window(self) -> deque:
        """Window queue property.

        :return: window queue
        :rtype: deque
        """
        return self._additional_vars["window"]

    @window.setter
    def window(self, value: deque) -> None:
        """Window queue setter.

        :param value: value to be set
        :type value: deque
        :raises TypeError: Type error exception
        """
        if not isinstance(value, deque):
            raise TypeError("value must be of type deque.")
        self._additional_vars["window"] = value

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1

        self.window.append(value)

        window_size = len(self.window)
        if window_size >= self.config.min_num_instances:
            # fmt: off
            num_first_samples = window_size - self.config.num_test_instances  # type: ignore # noqa: E501 pylint: disable=line-too-long
            r_samples = [*itertools.islice(self.window,
                                           num_first_samples,
                                           window_size)]
            w_samples = np.random.choice(
                a=[*itertools.islice(self.window, 0, num_first_samples)],
                size=self.config.num_test_instances,  # type: ignore
                replace=False,
            )
            # fmt: on
            _, p_value = ks_2samp(
                data1=w_samples,
                data2=r_samples,
                alternative="two-sided",
                method="auto",
            )

            if p_value <= self.config.alpha:  # type: ignore
                # Drift detected
                self.drift = True
            else:
                self.drift = False
        else:
            self.drift = False

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.window.clear()



"""Data drift exception module."""


class DimensionError(Exception):
    """Dimension exception."""


class GetStatisticalTestError(Exception):
    """Get statistical test exception."""


class MismatchDimensionError(Exception):
    """Miss match dimension exception."""


class MissingFitError(Exception):
    """Missing fit exception."""


class InsufficientSamplesError(Exception):
    """Insufficient samples exception."""



"""Data structures module."""

from typing import Any, Optional, Union, Tuple

import numpy as np  # type: ignore


class EmptyQueueError(Exception):
    """Empty queue exception.

    :param args: exception arguments
    :type args: Tuple[Any, ...]
    :param msg: exception message, defaults to "Queue is empty."
    :type msg: str
    :param kwargs: exception keyword arguments
    :type kwargs: dict[str, dict[str, Any]]
    """

    def __init__(  # noqa: D107
        self,
        *args: Tuple[Any, ...],
        msg: str = "Queue is empty.",
        **kwargs: dict[str, dict[str, Any]],
    ) -> None:
        super().__init__(
            msg,
            *args,
            **kwargs,
        )


class CircularQueue:
    """Class representing a circular queue.

    :param max_len: maximum capacity
    :type max_len: int
    """

    def __init__(  # noqa: D107
        self,
        max_len: int,
    ) -> None:
        self.count = 0
        self.first = 0
        self.last = -1
        self.max_len = max_len
        self.queue = [None] * self.max_len

    @property
    def count(self) -> int:
        """Number of total elements property.

        :return: Number of total elements
        :rtype: int
        """
        return self._count

    @count.setter
    def count(self, value: int) -> None:
        """Number of total elements setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("count must be greater or equal than 0.")
        self._count = value

    @property
    def first(self) -> int:
        """First queue index property.

        :return: first queue index
        :rtype: int
        """
        return self._first

    @first.setter
    def first(self, value: int) -> None:
        """First queue index setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("first must be greater or equal than 0.")
        self._first = value

    @property
    def last(self) -> int:
        """Last queue index property.

        :return: last queue index
        :rtype: int
        """
        return self._last

    @last.setter
    def last(self, value: int) -> None:
        """Last queue index setter.

        :param value: value to be set
        :type value: int
        """
        self._last = value

    @property
    def max_len(self) -> int:
        """Maximum number of allowed elements property.

        :return: maximum number of allowed elements
        :rtype: int
        """
        return self._max_len

    @max_len.setter
    def max_len(self, value: int) -> None:
        """Maximum number of allowed elements setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("max_len must be greater or equal than 0.")
        self._max_len = value

    @property
    def queue(self) -> list[Optional[Any]]:
        """Queue property.

        :return: queue
        :rtype: list[Optional[Any]]
        """
        return self._queue

    @queue.setter
    def queue(self, value: list[Optional[Any]]) -> None:
        """Queue setter.

        :param value: value to be set
        :type value: list[Optional[Any]]
        :raises ValueError: Value error exception
        """
        if not isinstance(value, list):
            raise TypeError("queue must be of type list.")
        self._queue = value

    @property
    def size(self) -> int:
        """Number of current elements property.

        :return: Number of current elements
        :rtype: int
        """
        return self.count

    def clear(self) -> None:
        """Clear queue."""
        self.count = 0
        self.first = 0
        self.last = -1
        self.queue = [None] * self.max_len

    def dequeue(self) -> Any:
        """Dequeue oldest element.

        :rtype: value: Any
        :raises EmptyQueue: Empty queue error exception
        """
        if self.is_empty():
            raise EmptyQueueError()
        element = self.queue[self.first]
        self.first = (self.first + 1) % self.max_len
        self.count -= 1
        return element  # type: ignore

    def enqueue(self, value: Union[np.ndarray, int, float]) -> Optional[Any]:
        """Enqueue element/s.

        :param value: value to be enqueued
        :type value: Union[np.ndarray, float]
        :return element: dequeued element
        :rtype: Optional[Any]
        """
        element = self.dequeue() if self.is_full() else None
        self.last = (self.last + 1) % self.max_len
        self.queue[self.last] = value  # type: ignore
        self.count += 1
        return element

    def is_empty(self) -> bool:
        """Check if queue is empty.

        :return: check if queue is empty
        :rtype: bool
        """
        return self.size == 0

    def is_full(self) -> bool:
        """Check if queue is full.

        :return: check if queue is full
        :rtype: bool
        """
        return self.size == self.max_len

    def __len__(self) -> int:
        """Queue size.

        :return: queue size
        :rtype: int
        """
        return self.size

    def maintain_last_element(self) -> None:
        """Clear all elements except the last one."""
        self.first = self.last
        self.count = 1

    def __getitem__(self, idx: int) -> Any:
        """Get queue item by position.

        :param idx: position index
        :type idx: int
        :return: queue item
        :rtype: Any
        """
        return self.queue[idx]  # type: ignore


class AccuracyQueue(CircularQueue):
    """Class representing an accuracy queue.

    :param max_len: maximum capacity
    :type max_len: int
    """

    def __init__(  # noqa: D107
        self,
        max_len: int,
    ) -> None:
        super().__init__(max_len=max_len)
        self.num_true = 0

    @property
    def num_false(self):
        """Number of false label property.

        :return: number of false labels
        :rtype: int
        """
        return self.count - self.num_true

    @property
    def num_true(self) -> int:
        """Number of true label property.

        :return: number of true labels
        :rtype: int
        """
        return self._num_true

    @num_true.setter
    def num_true(self, value: int) -> None:
        """Number of true labels setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("num_true value must be greater or equal than 0.")
        self._num_true = value

    def clear(self) -> None:
        """Clear queue."""
        super().clear()
        self.num_true = 0

    def dequeue(self) -> bool:
        """Dequeue oldest element.

        :return oldest element
        :rtype: bool
        :raises EmptyQueue: Empty queue error exception
        """
        element = super().dequeue()
        self.num_true -= 1 if element else 0
        return element

    def enqueue(self, value: Union[np.ndarray, float]) -> None:
        """Enqueue element/s.

        :param value: value to be enqueued
        :type value: Union[np.ndarray, float]
        """
        super().enqueue(value=value)
        self.num_true += np.count_nonzero(value)


# FIXME: There seem to be a bug on the treap DS. Uncomment all  # pylint: disable=fixme
#  the commented code lines when that is solved.
# class Node:
#     """Class representing a node of a treap."""
#
#     def __init__(
#         self,
#         key: Tuple[float, Union[int, float]],
#         value: Union[int, float],
#     ) -> None:
#         """Init method.
#
#         :param key: key
#         :type key: Tuple[float, Union[int, float]]
#         :param value: value
#         :type value: Union[int, float]]
#         """
#         self.key = key
#         self.value_ = value
#         self.priority = np.random.rand()
#         self.size = 1
#         self.height = 1
#         self.lazy = 0
#         self.max_value = value
#         self.min_value = value
#         self.left = None
#         self.right = None
#
#     @property
#     def key(self) -> Tuple[float, Union[int, float]]:
#         """Key property.
#
#         :return: key value
#         :rtype: Tuple[float, Union[int, float]]
#         """
#         return self._key
#
#     @key.setter
#     def key(self, value: Tuple[float, Union[int, float]]) -> None:
#         """Key setter.
#
#         :param value: value to be set
#         :type value: int
#         :raises TypeError: Type error exception
#         """
#         if not (isinstance(value[0], float) and isinstance(value[1], (int, float))):
#             raise TypeError("value must be a tuple of float and int or float.")
#         self._key = value
#
#     @property
#     def value_(self) -> Union[int, float]:
#         """Value property.
#
#         :return: value
#         :rtype: Union[int, float]
#         """
#         return self._value
#
#     @value_.setter
#     def value_(self, value: Union[int, float]) -> None:
#         """Value setter.
#
#         :param value: value to be set
#         :type value: int
#         :raises TypeError: Type error exception
#         """
#         if not isinstance(value, (int, float)):
#             raise TypeError("value must be of type int or float.")
#         self._value = value
#
#     @property
#     def priority(self) -> float:
#         """Priority property.
#
#         :return: priority value
#         :rtype: float
#         """
#         return self._priority
#
#     @priority.setter
#     def priority(self, value: float) -> None:
#         """Value setter.
#
#         :param value: value to be set
#         :type value: float
#         :raises ValueError: Value error exception
#         """
#         if not 0.0 <= value <= 1.0:
#             raise ValueError("value must be in the range [0, 1].")
#         self._priority = value
#
#     @property
#     def size(self) -> int:
#         """Size property.
#
#         :return: size value
#         :rtype: int
#         """
#         return self._size
#
#     @size.setter
#     def size(self, value: int) -> None:
#         """Size setter.
#
#         :param value: value to be set
#         :type value: int
#         :raises ValueError: Value error exception
#         """
#         if value < 0:
#             raise ValueError("value must be a positive number.")
#         self._size = value
#
#     @property
#     def height(self) -> int:
#         """Height property.
#
#         :return: height value
#         :rtype: int
#         """
#         return self._height
#
#     @height.setter
#     def height(self, value: int) -> None:
#         """Height setter.
#
#         :param value: value to be set
#         :type value: int
#         :raises ValueError: Value error exception
#         """
#         if value < 0:
#             raise ValueError("value must be a positive number.")
#         self._height = value
#
#     @property
#     def lazy(self) -> Union[int, float]:
#         """Lazy property.
#
#         :return: lazy value
#         :rtype: Union[int, float]
#         """
#         return self._lazy
#
#     @lazy.setter
#     def lazy(self, value: Union[int, float]) -> None:
#         """Lazy setter.
#
#         :param value: value to be set
#         :type value: Union[int, float]
#         """
#         self._lazy = value
#
#     @property
#     def max_value(self) -> Union[int, float]:
#         """Maximum value property.
#
#         :return: maximum value
#         :rtype: Union[int, float]
#         """
#         return self._max_value
#
#     @max_value.setter
#     def max_value(self, value: Union[int, float]) -> None:
#         """Maximum setter.
#
#         :param value: value to be set
#         :type value: Union[int, float]
#         """
#         self._max_value = value
#
#     @property
#     def min_value(self) -> Union[int, float]:
#         """Minimum value property.
#
#         :return: minimum value
#         :rtype: Union[int, float]
#         """
#         return self._min_value
#
#     @min_value.setter
#     def min_value(self, value: Union[int, float]) -> None:
#         """Minimum setter.
#
#         :param value: value to be set
#         :type value: Union[int, float]
#         """
#         self._min_value = value
#
#     @property
#     def left(self):
#         """Left node property.
#
#         :return: left node
#         :rtype: Optional[Node]
#         """
#         return self._left
#
#     @left.setter
#     def left(self, value) -> None:
#         """Left node setter.
#
#         :param value: value to be set
#         :type value: Optional[Node]
#         """
#         self._left = value
#
#     @property
#     def right(self):
#         """Right node property.
#
#         :return: right node
#         :rtype: Optional[Node]
#         """
#         return self._right
#
#     @right.setter
#     def right(self, value) -> None:
#         """Right node setter.
#
#         :param value: value to be set
#         :type value: Optional[Node]
#         """
#         self._right = value
#
#     def increase(self, value: Union[int, float]) -> None:
#         """Increase node by a value.
#
#         :param value: value to use
#         :type value: Union[int, float]
#         """
#         self.value_ += value
#         self.max_value += value
#         self.min_value += value
#         self.lazy += value
#
#     def unlazy(self) -> None:
#         """Unlazy node."""
#         if self.left is not None:
#             self.left.increase(value=self.lazy)
#         if self.right is not None:
#             self.right.increase(value=self.lazy)
#         self.lazy = 0
#
#     def split_first(self):
#         """Split first element."""
#         self.unlazy()
#
#         if self.left is not None:
#             left, self.left = self.left.split_first()
#             right = self
#         else:
#             right = self.right
#             self.right = None
#             left = self
#
#         if self.left is not None:
#             self.left.update()
#         if self.right is not None:
#             self.right.update()
#         return left, right
#
#     def split_last(self):
#         """Split last element."""
#         self.unlazy()
#
#         if self.right is not None:
#             self.right, right = self.right.split_last()
#             left = self
#         else:
#             left = self.left
#             self.left = None
#             right = self
#
#         if self.left is not None:
#             self.left.update()
#         if self.right is not None:
#             self.right.update()
#         return left, right
#
#     def update(self):
#         """Update node values."""
#         self.unlazy()
#
#         self.size = 1
#         self.height = 0
#         self.max_value = self.value_
#         self.min_value = self.value_
#
#         if self.left is not None:
#             self.size += self.left.size
#             self.height = self.left.height
#             self.max_value = max(self.max_value, self.left.max_value)
#             self.min_value = min(self.min_value, self.left.min_value)
#
#         if self.right is not None:
#             self.size += self.right.size
#             self.height = max(self.height, self.right.height)
#             self.max_value = max(self.max_value, self.right.max_value)
#             self.min_value = min(self.min_value, self.right.min_value)
#
#         self.height += 1
#
#     def split(self, key: Tuple[float, Union[int, float]]):
#         """Split."""
#         self.unlazy()
#
#         if key <= self.key:
#             left, self.left = (
#                 self.left.split(key=key) if self.left is not None else (None, None)
#             )
#             right = self
#         else:
#             self.right, right = (
#                 self.right.split(key=key) if self.right is not None else (None, None)
#             )
#             left = self
#         if left is not None:
#             left.update()
#         if right is not None:
#             right.update()
#         return left, right
#
#
# class Treap:
#     """Class representing a treap."""
#
#     def __init__(self, r: float = 1.0) -> None:
#         """Init method.
#
#         :param r: constant that satisfies |A| = r|B|
#         :type r: float
#         """
#         self.root = None
#         self.r = r
#         self.num_samples = [0, 0]
#
#     @property
#     def root(self) -> Optional[Node]:
#         """Root node property.
#
#         :return: root node
#         :rtype: Optional[Node]
#         """
#         return self._root
#
#     @root.setter
#     def root(self, value: Optional[Node]) -> None:
#         """Root node setter.
#
#         :param value: value to be set
#         :type value: int
#         :raises TypeError: Type error exception
#         """
#         if value is not None and not isinstance(value, Node):
#             raise TypeError("value must be of type Node or None.")
#         self._root = value
#
#     @property
#     def r(self) -> Union[int, float]:
#         """R value property.
#
#         :return: r value
#         :rtype: Union[int, float]
#         """
#         return self._r
#
#     @r.setter
#     def r(self, value: Union[int, float]) -> None:
#         """R value setter.
#
#         :param value: value to be set
#         :type value: Union[int, float]
#         :raises TypeError: Type error exception
#         """
#         if not isinstance(value, (int, float)):
#             raise TypeError("value must be of type int or float.")
#         self._r = value
#
#     @property
#     def num_samples(self) -> list[int]:
#         """Number of samples property.
#
#         :return: number of samples
#         :rtype: list[int]
#         """
#         return self._num_samples
#
#     @num_samples.setter
#     def num_samples(self, value: list[int]) -> None:
#         """Number of samples value setter.
#
#         :param value: value to be set
#         :type value: list[int]
#         :raises ValueError: Value error exception
#         """
#         if value[0] < 0 and value[1] < 0:
#             raise ValueError("value must be a positive number in both elements.")
#         self._num_samples = value
#
#     @property
#     def max(self) -> Union[int, float]:
#         """Maximum value property.
#
#         :return: maximum value
#         :rtype: Union[int, float]
#         """
#         return self.root.max_value  # type: ignore
#
#     @property
#     def min(self) -> Union[int, float]:
#         """Minimum value property.
#
#         :return: minimum value
#         :rtype: Union[int, float]
#         """
#         return self.root.min_value  # type: ignore
#
#     def insert(self, obs: float, group: int) -> None:
#         """Insert an observation by group.
#
#         :param obs: observation to insert
#         :type obs: float
#         :param group: group to which the observations belongs
#         :type group: int
#         """
#         key = (obs, group)
#         r = 1 if group == 0 else -self.r
#
#         self.num_samples[group] += 1
#
#         left, right = self.root.split(key) if self.root is not None else (None, None)
#
#         left, temp = left.split_last() if left is not None else (None, None)
#         initial_value = 0 if temp is None else temp.value_
#
#         left = self.merge(left=left, right=temp)
#
#         right = self.merge(left=Node(key=key, value=initial_value), right=right)
#         right.increase(value=r)
#
#         self.root = self.merge(left=left, right=right)
#
#     def remove(self, obs: float, group: int) -> None:
#         """Remove an observation by group.
#
#         :param obs: observation to remove
#         :type obs: float
#         :param group: group to which the observations belongs
#         :type group: int
#         """
#         key = (obs, group)
#         r = -1 if group == 0 else self.r
#
#         self.num_samples[group] -= 1
#
#         left, right = self.root.split(key) if self.root is not None else (None, None)
#         temp, right = right.split_first() if right is not None else (None, None)
#
#         if right is not None and temp is not None and temp.key == key:
#             right.increase(value=r)
#         else:
#             right = self.merge(left=temp, right=right)
#
#         self.root = self.merge(left=left, right=right)
#
#     def merge(self, left: Optional[Node], right: Optional[Node]) -> Optional[Node]:
#         """Merge two given nodes.
#
#         :param left: left node
#         :type left: Optional[Node]
#         :param right: right node
#         :type right: Optional[Node]
#         :return result of merging the two nodes
#         :rtype Optional[Node]
#         """
#         if left is None or right is None:
#             return left or right
#
#         if left.priority > right.priority:
#             left.unlazy()
#             left.right = self.merge(left=left.right, right=right)
#             node = left
#         else:
#             right.unlazy()
#             right.left = self.merge(left=left, right=right.left)
#             node = right
#         node.update()
#         return node



"""Checks module."""

from typing import Any

from frouros.callbacks.base import BaseCallback


def check_callbacks(
    callbacks: Any,
    expected_cls: BaseCallback,
) -> None:
    """Check callbacks.

    :param callbacks: callbacks
    :type callbacks: Any
    :param expected_cls: expected callback class
    :type expected_cls: BaseCallback
    :raises TypeError: Type error exception
    """
    if not (
        callbacks is None
        or isinstance(callbacks, expected_cls)  # type: ignore
        or (
            isinstance(callbacks, list)
            and all(
                isinstance(item, expected_cls) for item in callbacks  # type: ignore
            )
        )
    ):
        raise TypeError(
            f"callbacks must be of type None, "
            f"{expected_cls.name} or a list of {expected_cls.name}."
        )



"""EDDM (Early drift detection method) module."""

import copy
from typing import Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.statistical_process_control.base import (
    BaseSPCConfig,
    BaseSPC,
)


class EDDMConfig(BaseSPCConfig):
    """EDDM (Early drift detection method) [baena2006early]_ configuration.

    :param alpha: warning zone value, defaults to 0.95
    :type alpha: float
    :param beta: change zone value, defaults to 0.9
    :type beta: float
    :param level: level factor, defaults to 2.0
    :type level: float
    :param min_num_misclassified_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_misclassified_instances: int

    :References:

    .. [baena2006early] Baena-Garcıa, Manuel, et al. "Early drift detection method."
        Fourth international workshop on knowledge discovery from data streams.
        Vol. 6. 2006.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        alpha: float = 0.95,
        beta: float = 0.9,
        level: float = 2.0,
        min_num_misclassified_instances: int = 30,
    ) -> None:
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.level = level
        self.min_num_misclassified_instances = min_num_misclassified_instances

    @property
    def alpha(self) -> float:
        """Alpha property.

        :return: warning zone value
        :rtype: float
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value: float) -> None:
        """Alpha setter.

        :param value: value to be set
        :type value: float
        """
        self._alpha = value

    @property
    def beta(self) -> float:
        """Beta property.

        :return: change zone value
        :rtype: float
        """
        return self._beta

    @beta.setter
    def beta(self, value: float) -> None:
        """Beta setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value <= 0.0:
            raise ValueError("beta must be greater than 0.0.")
        if value >= self.alpha:
            raise ValueError("beta must be less than alpha.")
        self._beta = value

    @property
    def level(self) -> float:
        """Level property.

        :return: Level to use in detecting drift
        :rtype: float
        """
        return self._level

    @level.setter
    def level(self, value: float) -> None:
        """Level setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value <= 0.0:
            raise ValueError("drift level must be greater than 0.0.")
        self._level = value

    @property
    def min_num_misclassified_instances(self) -> int:
        """Minimum number of misclassified instances property.

        :return: minimum number of misclassified instances to use
        :rtype: float
        """
        return self._min_num_misclassified_instances

    @min_num_misclassified_instances.setter
    def min_num_misclassified_instances(self, value: int) -> None:
        """Minimum number of misclassified instances setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError(
                "min_num_misclassified_instances must be greater or equal than 0."
            )
        self._min_num_misclassified_instances = value


class EDDM(BaseSPC):
    """EDDM (Early drift detection method) [baena2006early]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`EDDMConfig` is used.
    :type config: Optional[EDDMConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :Note:
    :func:`update` method expects to receive a value of 0 if the instance is correctly classified (no error) and 1 otherwise (error).

    :References:

    .. [baena2006early] Baena-Garcıa, Manuel, et al. "Early drift detection method."
        Fourth international workshop on knowledge discovery from data streams.
        Vol. 6. 2006.

    :Example:

    >>> from frouros.detectors.concept_drift import EDDM
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.binomial(n=1, p=0.6, size=1000)
    >>> dist_b = np.random.binomial(n=1, p=0.8, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = EDDM()
    >>> warning_flag = False
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    ...     if not warning_flag and detector.warning:
    ...         print(f"Warning detected at step {i}")
    ...         warning_flag = True
    Warning detected at step 39
    Change detected at step 1294
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = EDDMConfig  # type: ignore

    def __init__(  # noqa: D107
        self,
        config: Optional[EDDMConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        # mean_distance_error = 0.0
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        mean_distance_error = 0.0
        self.additional_vars = {
            "last_distance_error": 0.0,
            "max_distance_threshold": float("-inf"),
            "mean_distance_error": mean_distance_error,
            "num_misclassified_instances": 0,
            "old_mean_distance_error": copy.copy(mean_distance_error),
            "std_distance_error": 0.0,
            "variance_distance_error": 0.0,
            **self.additional_vars,  # type: ignore
        }
        self._set_additional_vars_callback()

    @property
    def last_distance_error(self) -> float:
        """Last distance error property.

        :return: last distance error
        :rtype: float
        """
        return self._additional_vars["last_distance_error"]

    @last_distance_error.setter
    def last_distance_error(self, value: float) -> None:
        """Last distance error setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("last_distance_error must be great or equal than 0.")
        self._additional_vars["last_distance_error"] = value

    @property
    def max_distance_threshold(self) -> float:
        """Maximum distance threshold property.

        :return: maximum distance threshold
        :rtype: float
        """
        return self._additional_vars["max_distance_threshold"]

    @max_distance_threshold.setter
    def max_distance_threshold(self, value: float) -> None:
        """Maximum distance threshold setter.

        :param value: value to be set
        :type value: float
        """
        self._additional_vars["max_distance_threshold"] = value

    @property
    def mean_distance_error(self) -> float:
        """Mean distance error property.

        :return: mean distance error
        :rtype: float
        """
        return self._additional_vars["mean_distance_error"]

    @mean_distance_error.setter
    def mean_distance_error(self, value: float) -> None:
        """Mean distance error property.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("mean_distance_error must be great or equal than 0.")
        self._additional_vars["mean_distance_error"] = value

    @property
    def num_misclassified_instances(self) -> int:
        """Minimum number of misclassified instances property.

        :return: minimum number of misclassified instances to use
        :rtype: float
        """
        return self._additional_vars["num_misclassified_instances"]

    @num_misclassified_instances.setter
    def num_misclassified_instances(self, value: int) -> None:
        """Minimum number of misclassified instances setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError(
                "num_misclassified_instances must be greater or equal than 0."
            )
        self._additional_vars["num_misclassified_instances"] = value

    @property
    def old_mean_distance_error(self) -> float:
        """Old mean distance error property.

        :return: old mean distance error
        :rtype: float
        """
        return self._additional_vars["old_mean_distance_error"]

    @old_mean_distance_error.setter
    def old_mean_distance_error(self, value: float) -> None:
        """Old mean distance error setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("old_mean_distance_error must be great or equal than 0.")
        self._additional_vars["old_mean_distance_error"] = value

    @property
    def std_distance_error(self) -> float:
        """Standard deviation distance error property.

        :return: standard deviation distance error
        :rtype: float
        """
        return self._additional_vars["std_distance_error"]

    @std_distance_error.setter
    def std_distance_error(self, value: float) -> None:
        """Standard deviation distance error setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("std_distance_error must be great or equal than 0.")
        self._additional_vars["std_distance_error"] = value

    @property
    def variance_distance_error(self) -> float:
        """Variance distance error property.

        :return: variance deviation distance error
        :rtype: float
        """
        return self._additional_vars["variance_distance_error"]

    @variance_distance_error.setter
    def variance_distance_error(self, value: float) -> None:
        """Variance distance error setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("variance must be great or equal than 0.")
        self._additional_vars["variance_distance_error"] = value

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1

        if value == 1:
            self.num_misclassified_instances += 1

            distance = self.num_instances - self.last_distance_error
            self.old_mean_distance_error = self.mean_distance_error
            self.mean_distance_error += (
                distance - self.mean_distance_error
            ) / self.num_misclassified_instances
            self.variance_distance_error += (distance - self.mean_distance_error) * (
                distance - self.old_mean_distance_error
            )
            self.std_distance_error = (
                np.sqrt(self.variance_distance_error / self.num_misclassified_instances)
                if self.num_misclassified_instances > 0
                else 0.0
            )
            self.last_distance_error = self.num_instances

            if (
                self.num_instances
                >= self.config.min_num_misclassified_instances  # type: ignore
            ):
                distance_threshold = (
                    self.mean_distance_error
                    + self.config.level * self.std_distance_error  # type: ignore
                )
                if distance_threshold > self.max_distance_threshold:
                    self.max_distance_threshold = distance_threshold
                    self.drift, self.warning = False, False
                elif (
                    self.num_misclassified_instances
                    >= self.config.min_num_misclassified_instances  # type: ignore
                ):
                    p = distance_threshold / self.max_distance_threshold
                    if p < self.config.beta:  # type: ignore
                        # Out-of-Control
                        self.drift = True
                        self.warning = False
                    else:
                        if p < self.config.alpha:  # type: ignore
                            # Warning
                            self.warning = True
                        else:
                            self.warning = False
                        self.drift = False
        else:
            self.drift, self.warning = False, False

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.last_distance_error = 0.0
        self.max_distance_threshold = float("-inf")
        self.mean_distance_error = 0.0
        self.num_misclassified_instances = 0
        self.old_mean_distance_error = copy.copy(self.mean_distance_error)
        self.std_distance_error = 0.0
        self.variance_distance_error = 0.0



"""Configuration file for the tests."""

from typing import Optional, Tuple

import numpy as np  # type: ignore
import pytest  # type: ignore
import sklearn.linear_model  # type: ignore # pylint: disable=import-error
import sklearn.pipeline  # type: ignore # pylint: disable=import-error
import sklearn.preprocessing  # type: ignore # pylint: disable=import-error

from frouros.datasets.real import Elec2
from frouros.datasets.synthetic import Dummy, SEA
from frouros.metrics import PrequentialError


# Elec2 fixtures
@pytest.fixture(scope="function")
def elec2_raw() -> Elec2:
    """Elec2 raw dataset.

    :return: Elec2 raw dataset
    :rtype: Elec2
    """
    dataset = Elec2()
    dataset.download()
    return dataset


@pytest.fixture(scope="function")
def elec2(elec2_raw: Elec2) -> np.ndarray:  # pylint: disable=redefined-outer-name
    """Elec2 dataset.

    :param elec2_raw: Elec2 raw dataset
    :type elec2_raw: Elec2
    :return: Elec2 dataset
    :rtype: np.ndarray
    """
    return elec2_raw.load()


# SEA fixtures
@pytest.fixture(scope="function")
def sea() -> SEA:
    """SEA dataset generator.

    :return: SEA dataset generator
    :rtype: SEA
    """
    generator = SEA()
    return generator


@pytest.fixture(scope="function")
def stream_drift() -> np.ndarray:
    """Stream with drift.

    :return: stream with drift
    :rtype: np.ndarray
    """
    np.random.seed(seed=31)

    dist_normal = np.random.normal(
        loc=0.0,
        scale=0.5,
        size=100,
    )
    dist_drift_1 = np.random.normal(
        loc=4.0,
        scale=0.2,
        size=100,
    )
    dist_drift_2 = np.random.normal(
        loc=6.0,
        scale=0.5,
        size=100,
    )

    return np.concatenate((dist_normal, dist_drift_1, dist_drift_2))


@pytest.fixture(scope="module", name="clf_dataset")
def classification_dataset() -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Classification dataset using SEA generator.

    :return: classification dataset
    :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
    """
    concept_samples = 200
    generator = SEA(seed=31)

    concepts = [
        generator.generate_dataset(
            block=block,
            noise=0.0,
            num_samples=concept_samples,
        )
        for block in [1, 4, 3, 2]
    ]

    X, y = [], []  # noqa: N806
    for concept in concepts:
        for X_sample, y_sample in concept:  # noqa: N806
            X.append(X_sample)
            y.append(y_sample)
    X = np.array(X)  # noqa: N806
    y = np.array(y)

    split_idx = concept_samples

    X_ref = X[:split_idx]  # noqa: N806
    y_ref = y[:split_idx]

    X_test = X[split_idx:]  # noqa: N806
    y_test = y[split_idx:]

    return X_ref, y_ref, X_test, y_test


@pytest.fixture(scope="module")
def categorical_dataset() -> Tuple[np.ndarray, np.ndarray]:
    """Categorical variables dataset.

    :return: dataset
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    X_ref = np.array(  # noqa: N806
        ["a", "a", "b", "a", "a", "b"],
        dtype=object,
    )
    X_test = np.array(  # noqa: N806
        ["b", "a", "c", "c", "c", "c"],
        dtype=object,
    )

    return X_ref, X_test


@pytest.fixture(scope="module")
def numerical_dataset() -> Tuple[np.ndarray, np.ndarray]:
    """Numerical variables dataset.

    :return: dataset
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    np.random.seed(seed=31)
    X_ref = np.random.normal(loc=[0, 5], scale=[1, 1], size=(10, 2))  # noqa: N806
    X_test = np.random.normal(loc=[0, 1], scale=[1, 0.5], size=(10, 2))  # noqa: N806

    return X_ref, X_test


@pytest.fixture(scope="module")
def elec2_dataset() -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Elec2 dataset.

    :return: dataset
    :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray]
    """
    elec2 = Elec2()  # pylint: disable=redefined-outer-name
    elec2.download()
    dataset_ = elec2.load()

    # Two comprehensions lists are faster than iterating over one
    # (Python doing Python´s things).
    X = np.array(  # noqa: N806
        [sample.tolist()[:-1] for sample in dataset_], dtype=np.float16
    )
    y = np.array([sample[-1] for sample in dataset_], dtype="str")

    idx = X.shape[0] // 2
    X_ref = X[:idx]  # noqa: N806
    y_ref = y[:idx]

    X_test = X[idx:, :]  # noqa: N806

    return X_ref, y_ref, X_test


def _generate_multivariate_normal_samples(
    mean: np.ndarray, cov: np.ndarray, num_samples: int = 100, seed: int = 31
) -> np.ndarray:
    np.random.seed(seed=seed)
    X = np.random.multivariate_normal(  # noqa: N806
        mean=mean, cov=cov, size=num_samples
    )
    return X


@pytest.fixture(scope="module")
def multivariate_distribution_p() -> Tuple[np.ndarray, np.ndarray]:
    """Multivariate distribution p.

    :return: mean and covariance matrix of distribution p
    :rtype: Tuple[numpy.ndarray, numpy.ndarray]
    """
    mean = np.ones(2)
    cov = 2 * np.eye(2)

    return mean, cov


@pytest.fixture(scope="module")
def multivariate_distribution_q() -> Tuple[np.ndarray, np.ndarray]:
    """Multivariate distribution q.

    :return: mean and covariance matrix of distribution q
    :rtype: Tuple[numpy.ndarray, numpy.ndarray]
    """
    mean = np.zeros(2)
    cov = 3 * np.eye(2) - 1

    return mean, cov


@pytest.fixture(scope="module")
def X_ref_multivariate(  # noqa: N802
    multivariate_distribution_p: Tuple[  # pylint: disable=redefined-outer-name
        np.ndarray, np.ndarray
    ]
) -> np.ndarray:
    """Reference multivariate data.

    :param multivariate_distribution_p: multivariate distribution p
    :type multivariate_distribution_p: Tuple[numpy.ndarray, numpy.ndarray]
    :return: reference multivariate data
    :rtype: numpy.ndarray
    """
    X = _generate_multivariate_normal_samples(  # noqa: N806
        *multivariate_distribution_p
    )

    return X


@pytest.fixture(scope="module")
def X_test_multivariate(  # noqa: N802
    multivariate_distribution_q: Tuple[  # pylint: disable=redefined-outer-name
        np.ndarray, np.ndarray
    ]
) -> np.ndarray:
    """Test multivariate data.

    :param multivariate_distribution_q: multivariate distribution p
    :type multivariate_distribution_q: Tuple[numpy.ndarray, numpy.ndarray]
    :return: test multivariate data
    :rtype: numpy.ndarray
    """
    X = _generate_multivariate_normal_samples(  # noqa: N806
        *multivariate_distribution_q
    )

    return X


@pytest.fixture(scope="module")
def univariate_distribution_p() -> Tuple[float, float]:
    """Univariate distribution p.

    :return: mean and standard deviation of distribution p
    :rtype: Tuple[float, float]
    """
    mean, std = 1, 1

    return mean, std


@pytest.fixture(scope="module")
def univariate_distribution_q() -> Tuple[float, float]:
    """Univariate distribution q.

    :return: mean and standard deviation of distribution q
    :rtype: Tuple[float, float]
    """
    mean, std = 5, 2

    return mean, std


def _generate_univariate_normal_samples(
    mean: float, std: float, num_samples: int = 100, seed: int = 31
) -> np.ndarray:
    np.random.seed(seed=seed)
    X = np.random.normal(loc=mean, scale=std, size=num_samples)  # noqa: N806
    return X


@pytest.fixture(scope="module")
def X_ref_univariate(  # noqa: N802
    univariate_distribution_p: Tuple[  # pylint: disable=redefined-outer-name
        float, float
    ]
) -> np.ndarray:
    """Reference univariate data.

    :param univariate_distribution_p: univariate distribution p
    :type univariate_distribution_p: Tuple[float, float]
    :return: reference univariate data
    :rtype: numpy.ndarray
    """
    X = _generate_univariate_normal_samples(*univariate_distribution_p)  # noqa: N806

    return X


@pytest.fixture(scope="module")
def X_test_univariate(  # noqa: N802
    univariate_distribution_q: Tuple[  # pylint: disable=redefined-outer-name
        float, float
    ]
) -> np.ndarray:
    """Test multivariate data.

    :param univariate_distribution_q: univariate distribution q
    :type univariate_distribution_q: Tuple[float, float]
    :return: test univariate data
    :rtype: numpy.ndarray
    """
    X = _generate_univariate_normal_samples(*univariate_distribution_q)  # noqa: N806

    return X


@pytest.fixture(scope="module")
def prequential_error():
    """Prequential error.

    :return: prequential error
    :rtype: PrequentialError
    """
    return PrequentialError()


class DummyClassificationModel:
    """Dummy classification model class."""

    def __init__(self, num_classes: int = 2, seed: Optional[int] = None) -> None:
        """Init method.

        :param num_classes: number of classes
        :type num_classes: int
        """
        self.num_classes = num_classes
        np.random.seed(seed=seed)

    def fit(self, X: np.ndarray, y: np.ndarray, *args, **kwargs):  # noqa: N803, W0613
        """Fit method.

        :param X: feature data
        :type X: numpy.ndarray
        :param y: target data
        :type y: numpy.ndarray
        :return: random class prediction
        :rtype: numpy.ndarray
        """
        _ = (X, y, args, kwargs)
        return self

    def predict(self, X: np.ndarray) -> np.ndarray:  # noqa: N803
        """Predict method.

        :param X: feature data
        :type X: numpy.ndarray
        :return: random class prediction
        :rtype: numpy.ndarray
        """
        prediction = np.random.randint(low=0, high=self.num_classes, size=X.shape[0])
        return prediction


@pytest.fixture(scope="module")
def train_prediction_normal(
    clf_dataset: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
) -> np.ndarray:
    """Train a model and use a test dataset to obtain the predictions.

    :param clf_dataset: dataset generated using SEA
    :type clf_dataset: Tuple[numpy.ndarray, numpy.ndarray,
    numpy.ndarray, numpy.ndarray]
    :return: test predictions from trained model
    :rtype: numpy.ndarray
    """
    X_ref, y_ref, X_test, _ = clf_dataset  # noqa: N806

    model = DummyClassificationModel(num_classes=len(np.unique(y_ref)), seed=31)
    model.fit(X=X_ref, y=y_ref)
    y_pred = model.predict(X=X_test)

    return y_pred


@pytest.fixture(scope="module", name="dataset_simple")
def concept_drift_dataset_simple() -> (
    Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]
):
    """Dataset with multiple concepts to induce concept drift.

    :return: dataset split in reference and test
    :rtype: Tuple[Tuple[numpy.ndarray, numpy.ndarray],
    Tuple[numpy.ndarray, numpy.ndarray]]
    """
    concept_samples = 100
    concept_types = [1, 1, 0, 1]

    generator = Dummy(seed=31)
    concepts = [
        generator.generate_dataset(
            class_=concept_type,
            num_samples=concept_samples,
        )
        for concept_type in concept_types
    ]

    X, y = [], []  # noqa: N806
    for concept in concepts:
        for X_sample, y_sample in concept:  # noqa: N806
            X.append(X_sample)
            y.append(y_sample)
    X = np.array(X)  # noqa: N806
    y = np.array(y)

    num_ref_concept_types = 1
    idx_ref = num_ref_concept_types * concept_samples
    X_ref, y_ref = X[:idx_ref], y[:idx_ref]  # noqa: N806
    X_test, y_test = X[idx_ref:], y[idx_ref:]  # noqa: N806

    return (X_ref, y_ref), (X_test, y_test)


@pytest.fixture(scope="module", name="model")
def concept_drift_model(
    dataset_simple: Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]],
) -> sklearn.pipeline.Pipeline:
    """Model used for concept drift.

    :param dataset_simple: dataset with concept drift
    :type dataset_simple: Tuple[Tuple[numpy.ndarray, numpy.ndarray],
    :return: trained model
    :rtype: sklearn.pipeline.Pipeline
    """
    (X_ref, y_ref), _ = dataset_simple  # noqa: N806

    pipeline = sklearn.pipeline.Pipeline(
        [
            ("scaler", sklearn.preprocessing.StandardScaler()),
            ("model", sklearn.linear_model.LogisticRegression()),
        ]
    )
    pipeline.fit(X=X_ref, y=y_ref)

    return pipeline


@pytest.fixture(scope="module", name="model_errors")
def concept_drift_model_errors_simple(
    dataset_simple: Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]],
    model: sklearn.pipeline.Pipeline,
) -> list[int]:
    """Compute model errors given a dataset with concept drift.

    :param dataset_simple: dataset with concept drift
    :type dataset_simple: Tuple[Tuple[numpy.ndarray, numpy.ndarray],
    Tuple[numpy.ndarray, numpy.ndarray]]
    :param model: trained model
    :type model: sklearn.pipeline.Pipeline
    :return: model errors
    :rtype: list[int]
    """
    _, (X_test, y_test) = dataset_simple  # noqa: N806

    y_test_pred = model.predict(X_test)
    error = (1 - y_test_pred == y_test).astype(int).tolist()

    return error



"""Test synthetic datasets module."""

from typing import Any

import pytest  # type: ignore

from frouros.datasets.exceptions import InvalidBlockError
from frouros.datasets.synthetic import SEA


# SEA tests
@pytest.mark.parametrize("seed", [-1, "a"])
def test_sea_invalid_seed_error(seed: Any) -> None:
    """Test SEA invalid seed error.

    :param seed: seed value
    :type seed: Any
    """
    with pytest.raises((TypeError, ValueError)):
        _ = SEA(seed=seed)


@pytest.mark.parametrize("block", [0, 5])
def test_sea_invalid_block_error(sea: SEA, block: int) -> None:
    """Test SEA invalid block error.

    :param sea: SEA generator
    :type sea: SEA
    :param block: block to generate samples from
    :type block: int
    """
    with pytest.raises(InvalidBlockError):
        sea.generate_dataset(block=block)


@pytest.mark.parametrize("noise", [-0.1, 1.1])
def test_sea_invalid_noise_error(sea: SEA, noise: float) -> None:
    """Test SEA invalid noise error.

    :param sea: SEA generator
    :type sea: SEA
    :param noise: ratio of samples with a noisy class
    :type noise: float
    """
    with pytest.raises(ValueError):
        sea.generate_dataset(block=1, noise=noise)


@pytest.mark.parametrize("num_samples", [-1, 0])
def test_sea_invalid_num_samples_error(sea: SEA, num_samples: int) -> None:
    """Test SEA invalid number of samples error.

    :param sea: SEA generator
    :type sea: SEA
    :param num_samples: number of samples to generate
    :type num_samples: int
    """
    with pytest.raises(ValueError):
        sea.generate_dataset(block=1, num_samples=num_samples)



"""Test data drift detectors."""

from typing import Tuple, Union

import numpy as np  # type: ignore
import pytest  # type: ignore
from scipy.stats import PermutationMethod  # type: ignore

from frouros.detectors.data_drift.batch import (
    AndersonDarlingTest,
    BWSTest,
    ChiSquareTest,
    CVMTest,
    KSTest,
    KuiperTest,
    MannWhitneyUTest,
    WelchTTest,
)
from frouros.detectors.data_drift.batch import (
    BhattacharyyaDistance,
    EMD,
    EnergyDistance,
    HellingerDistance,
    HINormalizedComplement,
    PSI,
    JS,
    KL,
    MMD,
)
from frouros.detectors.data_drift.batch.base import BaseDataDriftBatch
from frouros.detectors.data_drift.streaming import (  # noqa: N811
    IncrementalKSTest,
    MMD as MMDStreaming,
)


@pytest.mark.parametrize(
    "detector, expected_statistic, expected_p_value",
    [(ChiSquareTest(), 6.13333333, 0.04657615)],
)
def test_batch_distance_based_categorical(
    categorical_dataset: Tuple[np.ndarray, np.ndarray],
    detector: ChiSquareTest,
    expected_statistic: float,
    expected_p_value: float,
) -> None:
    """Test batch categorical features method.

    :param categorical_dataset: categorical dataset
    :type categorical_dataset: Tuple[numpy.ndarray, numpy.ndarray]
    :param detector: detector test
    :type detector: ChiSquaredTest
    :param expected_statistic: expected statistic value
    :type expected_statistic: float
    :param expected_p_value: expected p-value value
    :type expected_p_value: float
    """
    X_ref, X_test = categorical_dataset  # noqa: N806

    _ = detector.fit(X=X_ref)
    (statistic, p_value), _ = detector.compare(X=X_test)

    assert np.isclose(statistic, expected_statistic)
    assert np.isclose(p_value, expected_p_value)


@pytest.mark.parametrize(
    "detector, expected_distance",
    [
        (EMD(), 3.85346006),
        (EnergyDistance(), 2.11059982),
        (JS(), 0.67010107),
        (KL(), np.inf),
        (HINormalizedComplement(), 0.78),
    ],
)
def test_batch_distance_based_univariate(
    X_ref_univariate: np.ndarray,  # noqa: N803
    X_test_univariate: np.ndarray,  # noqa: N803
    detector: BaseDataDriftBatch,
    expected_distance: float,
) -> None:
    """Test batch distance based univariate method.

    :param X_ref_univariate: reference univariate data
    :type X_ref_univariate: numpy.ndarray
    :param X_test_univariate: test univariate data
    :type X_test_univariate: numpy.ndarray
    :param detector: detector distance
    :type detector: BaseDataDriftBatch
    :param expected_distance: expected p-value value
    :type expected_distance: float
    """
    _ = detector.fit(X=X_ref_univariate)
    distance, _ = detector.compare(X=X_test_univariate)

    assert np.isclose(distance, expected_distance)


@pytest.mark.parametrize(
    "detector, expected_distance",
    [
        (PSI(), 461.20379435),
        (HellingerDistance(), 0.74509099),
        (BhattacharyyaDistance(), 0.55516059),
    ],
)
def test_batch_distance_bins_based_univariate_different_distribution(
    X_ref_univariate: np.ndarray,  # noqa: N803
    X_test_univariate: np.ndarray,  # noqa: N803
    detector: BaseDataDriftBatch,
    expected_distance: float,
) -> None:
    """Test batch distance bins based univariate different distribution method.

    :param X_ref_univariate: reference univariate data
    :type X_ref_univariate: numpy.ndarray
    :param X_test_univariate: test univariate data
    :type X_test_univariate: numpy.ndarray
    :param detector: detector distance
    :type detector: BaseDataDriftBatch
    :param expected_distance: expected p-value value
    :type expected_distance: float
    """
    _ = detector.fit(X=X_ref_univariate)
    distance, _ = detector.compare(X=X_test_univariate)

    assert np.isclose(distance, expected_distance)


@pytest.mark.parametrize(
    "detector, expected_distance",
    [
        (PSI(), 0.01840072),
        (HellingerDistance(), 0.04792538),
        (BhattacharyyaDistance(), 0.00229684),
    ],
)
def test_batch_distance_bins_based_univariate_same_distribution(
    univariate_distribution_p: Tuple[float, float],
    detector: BaseDataDriftBatch,
    expected_distance: float,
    num_samples: int = 500,
) -> None:
    """Test batch distance based univariate same distribution method.

    :param univariate_distribution_p: mean and standard deviation of distribution p
    :type univariate_distribution_p: Tuple[float, float]
    :param detector: detector distance
    :type detector: BaseDataDriftBatch
    :param expected_distance: expected p-value value
    :type expected_distance: float
    """
    np.random.seed(seed=31)
    X_ref = np.random.normal(*univariate_distribution_p, size=num_samples)  # noqa: N806
    X_test = np.random.normal(  # noqa: N806
        *univariate_distribution_p, size=num_samples
    )

    _ = detector.fit(X=X_ref)
    distance, _ = detector.compare(X=X_test)

    assert np.isclose(distance, expected_distance)


@pytest.mark.parametrize(
    "detector, expected_statistic, expected_p_value, kwargs",
    [
        (AndersonDarlingTest(), 23171.19994366, 0.001, {}),
        (
            BWSTest(),
            108757.63520694,
            0.00990099,
            {
                "method": PermutationMethod(
                    n_resamples=100,
                    random_state=31,
                ),
            },
        ),
        (CVMTest(), 3776.09848103, 5.38105056e-07, {}),
        (KSTest(), 0.99576271, 0.0, {}),
        (KuiperTest(), 0.99576271, 0.0, {}),
        (MannWhitneyUTest(), 6912.0, 0.0, {}),
        (WelchTTest(), -287.92032554, 0.0, {}),
    ],
)
def test_batch_statistical_univariate(
    elec2_dataset: Tuple[np.ndarray, np.ndarray, np.ndarray],
    detector: BaseDataDriftBatch,
    expected_statistic: float,
    expected_p_value: float,
    kwargs: dict,
) -> None:
    """Test statistical univariate method.

    :param elec2_dataset: Elec2 raw dataset
    :type elec2_dataset: Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
    :param detector: detector test
    :type detector: BaseDataDriftBatch
    :param expected_statistic: expected statistic value
    :type expected_statistic: float
    :param expected_p_value: expected p-value value
    :type expected_p_value: float
    :param kwargs: additional arguments
    :type kwargs: dict
    """
    X_ref, _, X_test = elec2_dataset  # noqa: N806

    _ = detector.fit(X=X_ref[:, 0])
    (statistic, p_value), _ = detector.compare(X=X_test[:, 0], **kwargs)

    assert np.isclose(statistic, expected_statistic)
    assert np.isclose(p_value, expected_p_value)


@pytest.mark.parametrize("detector, expected_distance", [(MMD(), 0.10163633)])
def test_batch_distance_based_multivariate_different_distribution(
    X_ref_multivariate: np.ndarray,  # noqa: N803
    X_test_multivariate: np.ndarray,  # noqa: N803
    detector: BaseDataDriftBatch,
    expected_distance: float,
) -> None:
    """Test batch distance based multivariate different distribution method.

    :param X_ref_multivariate: reference multivariate data
    :type X_ref_multivariate: numpy.ndarray
    :param X_test_multivariate: test multivariate data
    :type X_test_multivariate: numpy.ndarray
    :param detector: detector test
    :type detector: BaseDataDriftBatch
    :param expected_distance: expected distance value
    :type expected_distance: float
    """
    _ = detector.fit(X=X_ref_multivariate)
    statistic, _ = detector.compare(X=X_test_multivariate)

    assert np.isclose(statistic, expected_distance)


@pytest.mark.parametrize("detector, expected_distance", [(MMD(), 0.01570397)])
def test_batch_distance_based_multivariate_same_distribution(
    multivariate_distribution_p: Tuple[np.ndarray, np.ndarray],
    detector: BaseDataDriftBatch,
    expected_distance: float,
    num_samples: int = 100,
) -> None:
    """Test batch distance based multivariate same distribution method.

    :param multivariate_distribution_p: mean and covariance matrix of distribution p
    :type multivariate_distribution_p: Tuple[numpy.ndarray, numpy.ndarray]
    :param detector: detector test
    :type detector: BaseDataDriftBatch
    :param num_samples: number of random samples
    :type num_samples: int
    :param expected_distance: expected distance value
    :type expected_distance: float
    """
    np.random.seed(seed=31)
    X_ref = np.random.multivariate_normal(  # noqa: N806
        *multivariate_distribution_p, size=num_samples
    )
    X_test = np.random.multivariate_normal(  # noqa: N806
        *multivariate_distribution_p, size=num_samples
    )

    _ = detector.fit(X=X_ref)
    statistic, _ = detector.compare(X=X_test)

    assert np.isclose(statistic, expected_distance)


@pytest.mark.parametrize(
    "detector, expected_distance",
    [(MMD(chunk_size=10), 0.10163633), (MMD(chunk_size=None), 0.10163633)],
)
def test_batch_distance_based_chunk_size_valid(
    X_ref_multivariate: np.ndarray,  # noqa: N803
    X_test_multivariate: np.ndarray,  # noqa: N803
    detector: BaseDataDriftBatch,
    expected_distance: float,
) -> None:
    """Test batch distance based chunk size valid method.

    :param X_ref_multivariate: reference multivariate data
    :type X_ref_multivariate: numpy.ndarray
    :param X_test_multivariate: test multivariate data
    :type X_test_multivariate: numpy.ndarray
    :param detector: detector test
    :type detector: BaseDataDriftBatch
    :param expected_distance: expected distance value
    :type expected_distance: float
    """
    _ = detector.fit(X=X_ref_multivariate)
    statistic, _ = detector.compare(X=X_test_multivariate)

    assert np.isclose(statistic.distance, expected_distance)


@pytest.mark.parametrize(
    "chunk_size, expected_exception",
    [
        (1.5, TypeError),
        ("10", TypeError),
        (-1, ValueError),
    ],
)
def test_batch_distance_based_chunk_size_invalid(
    chunk_size: Union[int, float, str],
    expected_exception: Union[TypeError, ValueError],
) -> None:
    """Test batch distance based chunk size invalid method.

    :param chunk_size: chunk size
    :type chunk_size: Union[int, float, str]
    :param expected_exception: expected exception
    :type expected_exception: Union[TypeError, ValueError]
    """
    with pytest.raises(expected_exception):
        _ = MMD(chunk_size=chunk_size)  # type: ignore


@pytest.mark.parametrize(
    "detector, expected_statistic, expected_p_value",
    [
        (IncrementalKSTest(window_size=10), 0.27, 0.46046910),
    ],
)
def test_streaming_statistical_univariate_same_distribution(
    univariate_distribution_p: Tuple[float, float],
    detector: BaseDataDriftBatch,
    expected_statistic: float,
    expected_p_value: float,
) -> None:
    """Test streaming statistical test univariate same distribution method.

    :param univariate_distribution_p: mean and standard deviation of distribution p
    :type univariate_distribution_p: Tuple[float, float]
    :param detector: detector statistical test
    :type detector: BaseDataDriftStreaming
    :param expected_statistic: expected statistic value
    :type expected_statistic: float
    :param expected_p_value: expected p-value
    :type expected_p_value: float
    """
    np.random.seed(seed=31)
    X_ref = np.random.normal(*univariate_distribution_p, size=100)  # noqa: N806
    X_test = np.random.normal(*univariate_distribution_p, size=100)  # noqa: N806

    _ = detector.fit(X=X_ref)

    for value in X_test:
        test, _ = detector.update(value=value)  # type: ignore

    # Check last statistic and p-value
    assert np.isclose(test.statistic, expected_statistic)
    assert np.isclose(test.p_value, expected_p_value)


@pytest.mark.parametrize(
    "detector, expected_statistic, expected_p_value",
    [
        (IncrementalKSTest(window_size=10), 1.0, 0.0),
    ],
)
def test_streaming_statistical_univariate_different_distribution(
    univariate_distribution_p: Tuple[float, float],
    univariate_distribution_q: Tuple[float, float],
    detector: BaseDataDriftBatch,
    expected_statistic: float,
    expected_p_value: float,
) -> None:
    """Test streaming statistical test univariate different distribution method.

    :param univariate_distribution_p: mean and standard deviation of distribution p
    :type univariate_distribution_p: Tuple[float, float]
    :param univariate_distribution_q: mean and standard deviation of distribution q
    :type univariate_distribution_q: Tuple[float, float]
    :param detector: detector statistical test
    :type detector: BaseDataDriftStreaming
    :param expected_statistic: expected statistic value
    :type expected_statistic: float
    :param expected_p_value: expected p-value
    :type expected_p_value: float
    """
    np.random.seed(seed=31)
    X_ref = np.random.normal(*univariate_distribution_p, size=100)  # noqa: N806
    X_test = np.random.normal(*univariate_distribution_q, size=100)  # noqa: N806

    _ = detector.fit(X=X_ref)

    for value in X_test:
        test, _ = detector.update(value=value)  # type: ignore

    # Check last statistic and p-value
    assert np.isclose(test.statistic, expected_statistic)
    assert np.isclose(test.p_value, expected_p_value)


@pytest.mark.parametrize(
    "detector, expected_distance",
    [
        (MMDStreaming(window_size=10), -0.02327412),
    ],
)
def test_streaming_distance_based_univariate_same_distribution(
    univariate_distribution_p: Tuple[float, float],
    detector: BaseDataDriftBatch,
    expected_distance: float,
) -> None:
    """Test streaming distance based univariate same distribution method.

    :param univariate_distribution_p: mean and standard deviation of distribution p
    :type univariate_distribution_p: Tuple[float, float]
    :param detector: detector statistical test
    :type detector: BaseDataDriftStreaming
    :param expected_distance: expected distance value
    :type expected_distance: float
    """
    np.random.seed(seed=31)
    X_ref = np.random.normal(*univariate_distribution_p, size=100)  # noqa: N806
    X_test = np.random.normal(*univariate_distribution_p, size=100)  # noqa: N806

    _ = detector.fit(X=X_ref)

    for value in X_test:
        result, _ = detector.update(value=value)  # type: ignore

    # Check last distance
    assert np.isclose(result.distance, expected_distance)


@pytest.mark.parametrize(
    "detector, expected_distance",
    [
        (MMDStreaming(window_size=10), 0.8656735),
    ],
)
def test_streaming_distance_based_univariate_different_distribution(
    univariate_distribution_p: Tuple[float, float],
    univariate_distribution_q: Tuple[float, float],
    detector: BaseDataDriftBatch,
    expected_distance: float,
) -> None:
    """Test streaming distance based univariate different distribution method.

    :param univariate_distribution_p: mean and standard deviation of distribution p
    :type univariate_distribution_p: Tuple[float, float]
    :param univariate_distribution_q: mean and standard deviation of distribution q
    :type univariate_distribution_q: Tuple[float, float]
    :param detector: detector statistical test
    :type detector: BaseDataDriftStreaming
    :param expected_distance: expected distance value
    :type expected_distance: float
    """
    np.random.seed(seed=31)
    X_ref = np.random.normal(*univariate_distribution_p, size=100)  # noqa: N806
    X_test = np.random.normal(*univariate_distribution_q, size=100)  # noqa: N806

    _ = detector.fit(X=X_ref)

    for value in X_test:
        result, _ = detector.update(value=value)  # type: ignore

    # Check last distance
    assert np.isclose(result.distance, expected_distance)



"""Base data drift batch distance based module."""

import abc
from collections import namedtuple
from typing import Any, Callable, Optional, Tuple, Union

import numpy as np  # type: ignore
from scipy.stats import rv_histogram  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import (
    NumericalData,
    BaseStatisticalType,
    UnivariateData,
)
from frouros.detectors.data_drift.batch.base import BaseDataDriftBatch

DistanceResult = namedtuple("DistanceResult", ["distance"])


class BaseDistanceBased(BaseDataDriftBatch):
    """Abstract class representing a distance based detector."""

    def __init__(
        self,
        statistical_type: BaseStatisticalType,
        statistical_method: Callable,
        statistical_kwargs: dict[str, Any],
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        """Init method.

        :param statistical_type: statistical type
        :type statistical_type: BaseStatisticalType
        :param statistical_method: statistical method
        :type statistical_method: Callable
        :param statistical_kwargs: statistical kwargs
        :type statistical_kwargs: dict[str, Any]
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]
        """
        super().__init__(
            data_type=NumericalData(),
            statistical_type=statistical_type,
            callbacks=callbacks,
        )
        self.statistical_method = statistical_method
        self.statistical_kwargs = statistical_kwargs

    @property
    def statistical_method(self) -> Callable:
        """Statistical method property.

        :return: statistical method
        :rtype: Callable
        """
        return self._statistical_method

    @statistical_method.setter
    def statistical_method(self, value: Callable) -> None:
        """Statistical method setter.

        :param value: value to be set
        :type value: Callable
        :raises TypeError: Type error exception
        """
        if not isinstance(value, Callable):  # type: ignore
            raise TypeError("value must be of type Callable.")
        self._statistical_method = value

    @property
    def statistical_kwargs(self) -> dict[str, Any]:
        """Statistical kwargs property.

        :return: statistical kwargs
        :rtype: dict[str, Any]
        """
        return self._statistical_kwargs

    @statistical_kwargs.setter
    def statistical_kwargs(self, value: dict[str, Any]) -> None:
        """Statistical kwargs setter.

        :param value: value to be set
        :type value: dict[str, Any]
        """
        self._statistical_kwargs = value

    def _apply_method(
        self, X_ref: np.ndarray, X: np.ndarray, **kwargs  # noqa: N803
    ) -> DistanceResult:
        distance = self._distance_measure(X_ref=X_ref, X=X, **kwargs)
        return distance

    def _compare(
        self,
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        self._common_checks()  # noqa: N806
        self._specific_checks(X=X)  # noqa: N806
        distance = self._get_result(X=X, **kwargs)  # type: ignore
        return distance  # type: ignore

    @abc.abstractmethod
    def _distance_measure(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        pass


class BaseDistanceBasedBins(BaseDistanceBased):
    """Abstract class representing a distance based bins detector."""

    def __init__(
        self,
        statistical_method,
        statistical_kwargs,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
        num_bins: int = 10,
    ) -> None:
        """Init method.

        :param statistical_method: statistical method
        :type statistical_method: Callable
        :param statistical_kwargs: statistical kwargs
        :type statistical_kwargs: dict[str, Any]
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]
        :param num_bins: number of bins in which to divide probabilities
        :type num_bins: int
        """
        super().__init__(
            statistical_type=UnivariateData(),
            statistical_method=statistical_method,
            statistical_kwargs={**statistical_kwargs, "num_bins": num_bins},
            callbacks=callbacks,
        )
        self.num_bins = num_bins

    @property
    def num_bins(self) -> int:
        """Number of bins property.

        :return: number of bins in which to divide probabilities
        :rtype: int
        """
        return self._num_bins

    @num_bins.setter
    def num_bins(self, value: int) -> None:
        """Number of bins setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("value must be greater than 0.")
        self._num_bins = value

    def _distance_measure(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        distance_bins = self._distance_measure_bins(X_ref=X_ref, X=X)
        distance = DistanceResult(distance=distance_bins)
        return distance

    @staticmethod
    def _calculate_bins_values(
        X_ref: np.ndarray, X: np.ndarray, num_bins: int = 10  # noqa: N803
    ) -> np.ndarray:
        bins = np.histogram(np.hstack((X_ref, X)), bins=num_bins)[  # get the bin edges
            1
        ]
        X_ref_percents = (  # noqa: N806
            np.histogram(a=X_ref, bins=bins)[0] / X_ref.shape[0]
        )  # noqa: N806
        X_percents = np.histogram(a=X, bins=bins)[0] / X.shape[0]  # noqa: N806
        return X_ref_percents, X_percents

    @abc.abstractmethod
    def _distance_measure_bins(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
    ) -> float:
        pass


class BaseDistanceBasedProbability(BaseDistanceBased):
    """Abstract class representing a distance based probability detector."""

    def __init__(
        self,
        statistical_method,
        statistical_kwargs,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
        num_bins: int = 10,
    ) -> None:
        """Init method.

        :param statistical_method: statistical method
        :type statistical_method: Callable
        :param statistical_kwargs: statistical kwargs
        :type statistical_kwargs: dict[str, Any]
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]
        :param num_bins: number of bins in which to divide probabilities
        :type num_bins: int
        """
        super().__init__(
            statistical_type=UnivariateData(),
            statistical_method=statistical_method,
            statistical_kwargs=statistical_kwargs,
            callbacks=callbacks,
        )
        self.num_bins = num_bins

    @property
    def num_bins(self) -> int:
        """Number of bins property.

        :return: number of bins in which to divide probabilities
        :rtype: int
        """
        return self._num_bins

    @num_bins.setter
    def num_bins(self, value: int) -> None:
        """Number of bins setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("value must be greater than 0.")
        self._num_bins = value

    @abc.abstractmethod
    def _distance_measure(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        pass

    @staticmethod
    def _calculate_probabilities(
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
        num_bins: int,  # noqa: N803
    ) -> Tuple[np.ndarray, np.ndarray]:
        X_ref_rv_histogram = rv_histogram(  # noqa: N806
            np.histogram(X_ref, bins="auto")
        )
        X_rv_histogram = rv_histogram(np.histogram(X, bins="auto"))  # noqa: N806
        X_merge = np.concatenate([X_ref, X])  # noqa: N806
        bins = np.linspace(np.min(X_merge), np.max(X_merge), num_bins)
        X_ref_rvs = [  # noqa: N806
            X_ref_rv_histogram.cdf(bins[i])
            - X_ref_rv_histogram.cdf(bins[i - 1])  # noqa: N806
            for i in range(1, len(bins[1:]) + 1)
        ]
        X_rvs = [  # noqa: N806
            X_rv_histogram.cdf(bins[i]) - X_rv_histogram.cdf(bins[i - 1])  # noqa: N806
            for i in range(1, len(bins[1:]) + 1)
        ]
        return X_ref_rvs, X_rvs



"""KL (Kullback-Leibler divergence distance) module."""

from typing import Any, Optional, Union

import numpy as np  # type: ignore
from scipy.special import rel_entr  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBasedProbability,
    DistanceResult,
)


class KL(BaseDistanceBasedProbability):
    """KL (Kullback-Leibler divergence) [kullback1951information]_ detector.

    :param num_bins: number of bins in which to divide probabilities, defaults to 10
    :type num_bins: int
    :param callbacks: number of bins in which to divide probabilities, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]
    :param kwargs: additional keyword arguments to pass to scipy.special.rel_entr
    :type kwargs: dict[str, Any]

    :References:

    .. [kullback1951information] Kullback, Solomon, and Richard A. Leibler.
        "On information and sufficiency."
        The annals of mathematical statistics 22.1 (1951): 79-86.

    :Example:

    >>> from frouros.detectors.data_drift import KL
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = KL(num_bins=20)
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    DistanceResult(distance=inf)
    """

    def __init__(  # noqa: D107
        self,
        num_bins: int = 10,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            statistical_method=self._kl,
            statistical_kwargs={**kwargs, "num_bins": num_bins},
            callbacks=callbacks,
        )
        self.num_bins = num_bins
        self.kwargs = kwargs

    def _distance_measure(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        kl = self._kl(X=X_ref, Y=X, num_bins=self.num_bins, **self.kwargs)
        distance = DistanceResult(distance=kl)
        return distance

    @staticmethod
    def _kl(
        X: np.ndarray,  # noqa: N803
        Y: np.ndarray,
        *,
        num_bins: int,
        **kwargs: dict[str, Any],
    ) -> float:
        (  # noqa: N806
            X_ref_rvs,
            X_rvs,
        ) = BaseDistanceBasedProbability._calculate_probabilities(
            X_ref=X,
            X=Y,
            num_bins=num_bins,
        )
        kl = np.sum(rel_entr(X_rvs, X_ref_rvs, **kwargs))
        return kl



"""Test MMD."""

from functools import partial
from typing import Optional, Tuple

import numpy as np  # type: ignore
import pytest  # type: ignore

from frouros.detectors.data_drift import MMD
from frouros.utils.kernels import rbf_kernel


@pytest.mark.parametrize(
    "distribution_p, distribution_q, expected_distance",
    [
        ((0, 1, 100), (0, 1, 100), 0.00052755),  # (mean, std, size)
        ((0, 1, 100), (0, 1, 10), -0.03200193),
        ((0, 1, 10), (0, 1, 100), 0.07154671),
        ((2, 1, 100), (0, 1, 100), 0.43377622),
        ((2, 1, 100), (0, 1, 10), 0.23051378),
        ((2, 1, 10), (0, 1, 100), 0.62530767),
    ],
)
def test_mmd_batch_univariate(
    distribution_p: Tuple[float, float, int],
    distribution_q: Tuple[float, float, int],
    expected_distance: float,
) -> None:
    """Test MMD batch with univariate data.

    :param distribution_p: mean, std and size of samples from distribution p
    :type distribution_p: Tuple[float, float, int]
    :param distribution_q: mean, std and size of samples from distribution q
    :type distribution_q: Tuple[float, float, int]
    :param expected_distance: expected distance value
    :type expected_distance: float
    """
    np.random.seed(seed=31)
    X_ref = np.random.normal(*distribution_p)  # noqa: N806
    X_test = np.random.normal(*distribution_q)  # noqa: N806

    detector = MMD(
        kernel=partial(rbf_kernel, sigma=0.5),
    )
    _ = detector.fit(X=X_ref)

    result = detector.compare(X=X_test)[0]

    assert np.isclose(result.distance, expected_distance)


@pytest.mark.parametrize(
    "distribution_p, distribution_q, chunk_size",
    [
        ((0, 1, 100), (0, 1, 100), None),  # (mean, std, size)
        ((0, 1, 100), (0, 1, 100), 2),
        ((0, 1, 100), (0, 1, 100), 10),
        ((0, 1, 100), (0, 1, 10), None),
        ((0, 1, 100), (0, 1, 10), 2),
        ((0, 1, 100), (0, 1, 10), 10),
        ((0, 1, 10), (0, 1, 100), None),
        ((0, 1, 10), (0, 1, 100), 2),
        ((0, 1, 10), (0, 1, 100), 10),
    ],
)
def test_mmd_batch_precomputed_expected_k_xx(
    distribution_p: Tuple[float, float, int],
    distribution_q: Tuple[float, float, int],
    chunk_size: Optional[int],
) -> None:
    """Test MMD batch with precomputed expected k_xx.

    :param distribution_p: mean, std and size of samples from distribution p
    :type distribution_p: Tuple[float, float, int]
    :param distribution_q: mean, std and size of samples from distribution q
    :type distribution_q: Tuple[float, float, int]
    :param chunk_size: chunk size
    :type chunk_size: Optional[int]
    """
    np.random.seed(seed=31)
    X_ref = np.random.normal(*distribution_p)  # noqa: N806
    X_test = np.random.normal(*distribution_q)  # noqa: N806

    kernel = partial(rbf_kernel, sigma=0.5)

    detector = MMD(
        kernel=kernel,
        chunk_size=chunk_size,
    )
    _ = detector.fit(X=X_ref)

    # Computes mmd using precomputed expected k_xx
    precomputed_distance = detector.compare(X=X_test)[0].distance

    # Computes mmd from scratch
    scratch_distance = MMD._mmd(  # pylint: disable=protected-access
        X=X_ref,
        Y=X_test,
        kernel=kernel,
        chunk_size=chunk_size,
    )

    assert np.isclose(precomputed_distance, scratch_distance)



"""Test real datasets module."""

import numpy as np  # type: ignore
import pytest  # type: ignore

from frouros.datasets.exceptions import ReadFileError
from frouros.datasets.real import Elec2


# Elec2 tests
def test_elec2_file_not_found_error(elec2_raw: Elec2) -> None:
    """Test Elec2 file not found error.

    :param elec2_raw: Elec2 raw dataset
    :type elec2_raw: Elec2
    # :param elec2_delimiter: Elec2 load delimiter
    # :type elec2_delimiter: str
    """
    _ = elec2_raw.load()
    with pytest.raises(FileNotFoundError):
        _ = elec2_raw.load()


def test_elec2_permission_error() -> None:
    """Test Elec2 permission error."""
    with pytest.raises(PermissionError):
        Elec2(file_path="//elec2").download()


def test_elec2_read_file_error(elec2_raw: Elec2) -> None:
    """Test Elec2 read file error.

    :param elec2_raw: Elec2 raw dataset
    :type elec2_raw: Elec2
    """
    with pytest.raises(ReadFileError):
        _ = elec2_raw.load(index=2)


def test_elec2_shape(elec2: np.ndarray) -> None:
    """Test Elec2 shape.

    :param elec2: Elec2 dataset
    :type elec2: np.ndarray
    """
    assert elec2.shape == (45312,)


def test_elec2_type(elec2: np.ndarray) -> None:
    """Test Elec2 type.

    :param elec2: Elec2 dataset
    :type elec2: np.ndarray
    """
    assert isinstance(elec2, np.ndarray)



"""ECDD (EWMA for Concept Drift Detection) module."""

from typing import Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.statistical_process_control.base import (
    BaseSPC,
    BaseECDDConfig,
)
from frouros.utils.stats import EWMA, Mean


class ECDDWTConfig(BaseECDDConfig):
    """ECDDWT (EWMA Concept Drift Detection Warning) [ross2012exponentially]_ configuration.

    :param lambda_: weight given to recent data compared to older data, defaults to 0.2
    :type lambda_: float
    :param average_run_length: expected time between false positive detections [100, 400 or 1000], defaults to 400
    :type average_run_length: int
    :param warning_level: warning level value, defaults to 0.5
    :type warning_level: float
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_instances: int

    :References:

    .. [ross2012exponentially] Ross, Gordon J., et al.
        "Exponentially weighted moving average charts for detecting concept drift."
        Pattern recognition letters 33.2 (2012): 191-198.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        lambda_: float = 0.2,
        average_run_length: int = 400,
        warning_level: float = 0.5,
        min_num_instances: int = 30,
    ) -> None:
        super().__init__(
            lambda_=lambda_,
            average_run_length=average_run_length,
            warning_level=warning_level,
            min_num_instances=min_num_instances,
        )


class ECDDWT(BaseSPC):
    """ECDDWT (EWMA Concept Drift Detection Warning) [ross2012exponentially]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`ECDDWTConfig` is used.
    :type config: Optional[ECDDWTConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :Note:
    :func:`update` method expects to receive a value of 0 if the instance is correctly classified (no error) and 1 otherwise (error).

    :References:

    .. [ross2012exponentially] Ross, Gordon J., et al.
        "Exponentially weighted moving average charts for detecting concept drift."
        Pattern recognition letters 33.2 (2012): 191-198.

    :Example:

    >>> from frouros.detectors.concept_drift import ECDDWT
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.binomial(n=1, p=0.6, size=1000)
    >>> dist_b = np.random.binomial(n=1, p=0.8, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = ECDDWT()
    >>> warning_flag = False
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    ...     if not warning_flag and detector.warning:
    ...         print(f"Warning detected at step {i}")
    ...         warning_flag = True
    """  # noqa: E501 # pylint: disable=line-too-long

    config_type = ECDDWTConfig  # type: ignore

    def __init__(  # noqa: D107
        self,
        config: Optional[ECDDWTConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,  # type: ignore
            callbacks=callbacks,
        )
        self.additional_vars = {
            "p": Mean(),
            "z": EWMA(alpha=self.config.lambda_),  # type: ignore
            **self.additional_vars,  # type: ignore
        }
        self._set_additional_vars_callback()
        self._lambda_div_two_minus_lambda = self.config.lambda_ / (  # type: ignore
            2 - self.config.lambda_  # type: ignore
        )

    @property
    def p(self) -> Mean:
        """P property.

        :return: p value
        :rtype: Mean
        """
        return self._additional_vars["p"]

    @p.setter
    def p(self, value: Mean) -> None:
        """P setter.

        :param value: value to be set
        :type value: Mean
        :raises TypeError: Type error exception
        """
        if not isinstance(value, Mean):
            raise TypeError("value must be of type Mean.")
        self._additional_vars["p"] = value

    @property
    def z(self) -> EWMA:
        """Z property.

        :return: z value
        :rtype: Mean
        """
        return self._additional_vars["z"]

    @z.setter
    def z(self, value: EWMA) -> None:
        """Z setter.

        :param value: value to be set
        :type value: EWMA
        :raises TypeError: Type error exception
        """
        if not isinstance(value, EWMA):
            raise TypeError("value must be of type EWMA.")
        self._additional_vars["z"] = value

    def _check_threshold(
        self, control_limit: float, z_variance: float, warning_level: float = 1.0
    ) -> bool:
        return self.z.mean > self.p.mean + warning_level * control_limit * z_variance

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.p = Mean()
        self.z = EWMA(alpha=self.config.lambda_)  # type: ignore

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1

        self.p.update(value=value)
        self.z.update(value=value)

        if self.num_instances >= self.config.min_num_instances:
            error_rate_variance = self.p.mean * (1 - self.p.mean)
            z_variance = np.sqrt(
                self._lambda_div_two_minus_lambda
                * (1 - self.z.one_minus_alpha ** (2 * self.num_instances))
                * error_rate_variance
            )
            control_limit = self.config.control_limit_func(  # type: ignore
                p=self.p.mean,
            )

            drift_flag = self._check_threshold(
                control_limit=control_limit,
                z_variance=z_variance,
            )

            if drift_flag:
                # Out-of-Control
                self.drift = True
                self.warning = False
            else:
                warning_flag = self._check_threshold(
                    control_limit=control_limit,
                    z_variance=z_variance,
                    warning_level=self.config.warning_level,  # type: ignore
                )
                if warning_flag:
                    # Warning
                    self.warning = True
                else:
                    # In-Control
                    self.warning = False
                self.drift = False
        else:
            self.drift, self.warning = False, False



"""Base concept drift SPC (statistical process control) module."""

import abc
from typing import Optional, Tuple, Union

import numpy as np  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.exceptions import InvalidAverageRunLengthError
from frouros.detectors.concept_drift.streaming.base import (
    BaseConceptDriftStreamingConfig,
    BaseConceptDriftStreaming,
)
from frouros.utils.stats import Mean


class BaseSPCConfig(BaseConceptDriftStreamingConfig):
    """Class representing a SPC configuration class."""

    def __init__(
        self,
        warning_level: float = 2.0,
        drift_level: float = 3.0,
        min_num_instances: int = 30,
    ) -> None:
        """Init method.

        :param warning_level: warning level factor
        :type warning_level: float
        :param drift_level: drift level factor
        :type drift_level: float
        :param min_num_instances: minimum numbers of instances
        to start looking for changes
        :type min_num_instances: int
        """
        super().__init__(min_num_instances=min_num_instances)
        self.warning_level = warning_level
        self.drift_level = drift_level

    @property
    def drift_level(self) -> float:
        """Drift level property.

        :return: drift level to use in detecting drift
        :rtype: float
        """
        return self._drift_level  # type: ignore

    @drift_level.setter
    def drift_level(self, value: float) -> None:
        """Drift level setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value <= 0.0:
            raise ValueError("drift level must be greater than 0.0.")
        if value <= self.warning_level:
            raise ValueError("drift level must be greater than warning level.")
        self._drift_level = value

    @property
    def warning_level(self) -> float:
        """Warning level property.

        :return: warning level to use in detecting drift
        :rtype: float
        """
        return self._warning_level

    @warning_level.setter
    def warning_level(self, value: float) -> None:
        """Warning level setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value <= 0.0:
            raise ValueError("warning level must be greater than 0.0.")
        self._warning_level = value


class BaseSPC(BaseConceptDriftStreaming):
    """Abstract class representing an SPC estimator."""

    config_type = BaseSPCConfig

    def __init__(
        self,
        config: Optional[BaseSPCConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        """Init method.

        :param config: configuration parameters
        :type config: Optional[BaseSPCConfig]
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackStreaming,
        list[BaseCallbackStreaming]]]
        """
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        self.additional_vars = {
            "warning": False,
        }
        self._set_additional_vars_callback()

    @property
    def warning(self) -> bool:
        """Warning property.

        :return: warning value
        :rtype: bool
        """
        return self._additional_vars["warning"]

    @warning.setter
    def warning(self, value: bool) -> None:
        """Warning setter.

        :param value: value to be set
        :type value: bool
        """
        self._additional_vars["warning"] = value

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.warning = False

    @property
    def status(self) -> dict[str, bool]:
        """Status property.

        :return: status dict
        :rtype: dict[str, bool]
        """
        return {**super().status, "warning": self.warning}

    @abc.abstractmethod
    def _update(self, value: Union[int, float], **kwargs) -> None:
        pass


class BaseSPCError(BaseSPC):
    """Abstract class representing a SPC error estimator."""

    config_type = BaseSPCConfig

    def __init__(
        self,
        config: Optional[BaseSPCConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        """Init method.

        :param config: configuration parameters
        :type config: Optional[BaseSPCConfig]
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackStreaming,
        list[BaseCallbackStreaming]]]
        """
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        self.additional_vars = {
            "error_rate": Mean(),
            "min_error_rate": float("inf"),
            "min_std": float("inf"),
            **self.additional_vars,  # type: ignore
        }
        self._set_additional_vars_callback()

    @property
    def error_rate(self) -> Mean:
        """Error rate property.

        :return: error rate to use
        :rtype: Mean
        """
        return self._additional_vars["error_rate"]

    @error_rate.setter
    def error_rate(self, value: Mean) -> None:
        """Error rate setter.

        :param value: value to be set
        :type value: Mean
        """
        self._additional_vars["error_rate"] = value

    @property
    def min_error_rate(self) -> float:
        """Minimum error rate property.

        :return: minimum error rate to use
        :rtype: float
        """
        return self._additional_vars["min_error_rate"]

    @min_error_rate.setter
    def min_error_rate(self, value: float) -> None:
        """Minimum error rate setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("min_error_rate must be great or equal than 0.")
        self._additional_vars["min_error_rate"] = value

    @property
    def min_error_rate_plus_std(self) -> float:
        """Minimum error rate + std property.

        :return: minimum error rate + std to determine if a change is happening
        :rtype: float
        """
        return self.min_error_rate + self.min_std

    @property
    def min_std(self) -> float:
        """Minimum standard deviation property.

        :return: minimum standard deviation to use
        :rtype: float
        """
        return self._additional_vars["min_std"]

    @min_std.setter
    def min_std(self, value: float) -> None:
        """Minimum standard deviation setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value < 0:
            raise ValueError("min_std must be great or equal than 0.")
        self._additional_vars["min_std"] = value

    def _calculate_error_rate_plus_std(self) -> Tuple[float, float]:
        std = np.sqrt(
            self.error_rate.mean * (1 - self.error_rate.mean) / self.num_instances
        )
        error_rate_plus_std = self.error_rate.mean + std
        return error_rate_plus_std, std

    def _update_min_values(self, error_rate_plus_std: float, std: float) -> None:
        if error_rate_plus_std < self.min_error_rate_plus_std:
            self.min_error_rate = self.error_rate.mean
            self.min_std = std

    @staticmethod
    def _check_threshold(
        error_rate_plus_std: float, min_error_rate: float, min_std: float, level: float
    ) -> bool:
        return error_rate_plus_std > min_error_rate + level * min_std

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.error_rate = Mean()
        self.min_error_rate = float("inf")
        self.min_std = float("inf")

    @abc.abstractmethod
    def _update(self, value: Union[int, float], **kwargs) -> None:
        pass


class BaseECDDConfig(BaseConceptDriftStreamingConfig):
    """Class representing a ECDD configuration class."""

    average_run_length_map = {
        100: lambda p: 2.76
        - 6.23 * p
        + 18.12 * np.power(p, 3)
        - 312.45 * np.power(p, 5)
        + 1002.18 * np.power(p, 7),
        400: lambda p: 3.97
        - 6.56 * p
        + 48.73 * np.power(p, 3)
        - 330.13 * np.power(p, 5)
        + 848.18 * np.power(p, 7),
        1000: lambda p: 1.17
        + 7.56 * p
        - 21.24 * np.power(p, 3)
        + 112.12 * np.power(p, 5)
        - 987.23 * np.power(p, 7),
    }

    def __init__(
        self,
        lambda_: float = 0.2,
        average_run_length: int = 400,
        warning_level: float = 0.5,
        min_num_instances: int = 30,
    ) -> None:
        """Init method.

        :param average_run_length: expected time between false positive detections
        :type average_run_length: int
        :param lambda_: weight given to recent data compared to older data
        :type lambda_: float
        :param warning_level: warning level value
        :type warning_level: float
        :param min_num_instances: minimum numbers of instances
        to start looking for changes
        :type min_num_instances: int
        :raises InvalidAverageRunLengthError: Invalid average run length error exception
        """
        super().__init__(min_num_instances=min_num_instances)
        try:
            self.control_limit_func = self.average_run_length_map[average_run_length]
        except KeyError as e:
            raise InvalidAverageRunLengthError(
                "average_run_length must be 100, 400 or 1000."
            ) from e
        self.lambda_ = lambda_
        self.warning_level = warning_level

    @property
    def lambda_(self) -> float:
        """Weight recent data property.

        :return: weight given to recent data
        :rtype: float
        """
        return self._lambda_

    @lambda_.setter
    def lambda_(self, value: float) -> None:
        """Weight recent data setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if not 0.0 <= value <= 1.0:
            raise ValueError("lambda_ must be in the range [0, 1].")
        self._lambda_ = value

    @property
    def warning_level(self) -> float:
        """Warning level property.

        :return: warning level to use in detecting drift
        :rtype: float
        """
        return self._warning_level

    @warning_level.setter
    def warning_level(self, value: float) -> None:
        """Warning level setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if not 0.0 < value < 1.0:
            raise ValueError("warning level must be in the range (0.0, 1.0).")
        self._warning_level = value



"""Base callback module."""

import abc
from typing import Optional

import numpy as np  # type: ignore


class BaseCallback(abc.ABC):
    """Abstract class representing a callback."""

    def __init__(self, name: Optional[str] = None) -> None:
        """Init method.

        :param name: name value
        :type name: Optional[str]
        """
        self.name = name  # type: ignore
        self.detector = None
        self.logs = {}  # type: ignore

    @property
    def name(self) -> str:
        """Name property.

        :return: name value
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value: Optional[str]) -> None:
        """Name method setter.

        :param value: value to be set
        :type value: Optional[str]
        :raises TypeError: Type error exception
        """
        if not isinstance(value, str) and value is not None:
            raise TypeError("name must be of type str or None.")
        self._name = self.__class__.__name__ if value is None else value

    def set_detector(self, detector) -> None:
        """Set detector method."""
        self.detector = detector

    # @property
    # def detector(self) -> Optional[BaseConceptDrift, BaseDataDriftBatch]:
    #     return self._detector
    #
    # @detector.setter
    # def detector(self, value: Optional[BaseConceptDrift, BaseDataDriftBatch]) -> None:
    #     if not isinstance(
    #             value, (BaseConceptDrift, BaseDataDriftBatch)):
    #         raise TypeError(
    #             "value must be of type BaseConceptDrift or BaseDataDriftBatch."
    #         )
    #     self._detector = value

    def on_fit_start(self, X: np.ndarray) -> None:  # noqa: N803, B027
        """On fit start method.

        :param X: reference data
        :type X: numpy.ndarray
        """

    def on_fit_end(self, X: np.ndarray) -> None:  # noqa: N803, B027
        """On fit end method.

        :param X: reference data
        :type X: numpy.ndarray
        """

    @abc.abstractmethod
    def reset(self) -> None:
        """Reset method."""

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return f"{self.__class__.__name__}(name='{self.name}')"



"""DDM (Drift detection method) module."""

from contextlib import suppress
from typing import Union, Optional

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.statistical_process_control.base import (
    BaseSPCConfig,
    BaseSPCError,
)


class DDMConfig(BaseSPCConfig):
    """DDM (Drift detection method) [gama2004learning]_ configuration.

    :param warning_level: warning level factor, defaults to 2.0
    :type warning_level: float
    :param drift_level: drift level factor, defaults to 3.0
    :type drift_level: float
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_instances: int

    :References:

    .. [gama2004learning] Gama, Joao, et al.
        "Learning with drift detection."
        Advances in Artificial Intelligence–SBIA 2004: 17th Brazilian Symposium on
        Artificial Intelligence, Sao Luis, Maranhao, Brazil, September 29-October 1,
        2004. Proceedings 17. Springer Berlin Heidelberg, 2004.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        warning_level: float = 2.0,
        drift_level: float = 3.0,
        min_num_instances: int = 30,
    ) -> None:
        super().__init__(
            warning_level=warning_level,
            drift_level=drift_level,
            min_num_instances=min_num_instances,
        )


class DDM(BaseSPCError):
    """DDM (Drift detection method) [gama2004learning]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`DDMConfig` is used.
    :type config: Optional[DDMConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :Note:
    :func:`update` method expects to receive a value of 0 if the instance is correctly classified (no error) and 1 otherwise (error).

    :References:

    .. [gama2004learning] Gama, Joao, et al.
        "Learning with drift detection."
        Advances in Artificial Intelligence–SBIA 2004: 17th Brazilian Symposium on
        Artificial Intelligence, Sao Luis, Maranhao, Brazil, September 29-October 1,
        2004. Proceedings 17. Springer Berlin Heidelberg, 2004.

    :Example:

    >>> from frouros.detectors.concept_drift import DDM
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.binomial(n=1, p=0.6, size=1000)
    >>> dist_b = np.random.binomial(n=1, p=0.8, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = DDM()
    >>> warning_flag = False
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    ...     if not warning_flag and detector.warning:
    ...         print(f"Warning detected at step {i}")
    ...         warning_flag = True
    Warning detected at step 1049
    Change detected at step 1131
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = DDMConfig

    def __init__(  # noqa: D107
        self,
        config: Optional[DDMConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1
        self.error_rate.update(value=value)

        if self.num_instances >= self.config.min_num_instances:
            error_rate_plus_std, std = self._calculate_error_rate_plus_std()

            self._update_min_values(error_rate_plus_std=error_rate_plus_std, std=std)

            drift_flag = self._check_threshold(
                error_rate_plus_std=error_rate_plus_std,
                min_error_rate=self.min_error_rate,
                min_std=self.min_std,
                level=self.config.drift_level,  # type: ignore
            )
            if drift_flag:
                # Out-of-Control
                self.drift = True
                self.warning = False
            else:
                warning_flag = self._check_threshold(
                    error_rate_plus_std=error_rate_plus_std,
                    min_error_rate=self.min_error_rate,
                    min_std=self.min_std,
                    level=self.config.warning_level,  # type: ignore
                )
                if warning_flag:
                    # Warning
                    self.warning = True
                    for callback in self.callbacks:  # type: ignore
                        with suppress(AttributeError):
                            callback.on_warning_detected(**kwargs)  # type: ignore
                else:
                    # In-Control
                    self.warning = False
                self.drift = False
        else:
            self.drift, self.warning = False, False



"""Base data drift statistical test module."""

import abc
from collections import namedtuple
from typing import Tuple

import numpy as np  # type: ignore

from frouros.detectors.data_drift.batch.base import BaseDataDriftBatch

StatisticalResult = namedtuple("StatisticalResult", ["statistic", "p_value"])


class BaseStatisticalTest(BaseDataDriftBatch):
    """Abstract class representing a statistical test."""

    def _apply_method(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
        **kwargs,
    ) -> Tuple[float, float]:
        statistical_test = self._statistical_test(
            X_ref=X_ref,
            X=X,
            **kwargs,
        )
        return statistical_test

    def _compare(
        self,
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> StatisticalResult:
        self._common_checks()  # noqa: N806
        self._specific_checks(X=X)  # noqa: N806
        result = self._get_result(X=X, **kwargs)
        return result  # type: ignore

    @staticmethod
    @abc.abstractmethod
    def _statistical_test(
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
        **kwargs,
    ) -> StatisticalResult:
        pass



"""Base metrics module."""

import abc
from typing import Optional


class BaseMetric(abc.ABC):
    """Abstract class representing a metric."""

    def __init__(self, name: Optional[str] = None) -> None:
        """Init method.

        :param name: name value
        :type name: Optional[str]
        """
        self.name = type(self).__name__ if name is None else name

    @property
    def name(self) -> str:
        """Metrics' name property.

        :return: metrics' name
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Metrics' name setter.

        :param value: value to be set
        :type value: str
        :raises TypeError: Type error exception
        """
        if not isinstance(value, str):
            raise TypeError("value must be of type str.")
        self._name = value

    @abc.abstractmethod
    def __call__(
        self,
        error_value: float,
    ) -> float:
        """__call__ method that updates the metric error.

        :param error_value: error value
        :type error_value: float
        :return: cumulative error
        :rtype: Union[int, float]
        """

    @abc.abstractmethod
    def reset(self) -> None:
        """Reset method."""

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return f"{self.__class__.__name__}(name='{self.name}')"



"""Concept drift change detection methods' init."""

from .bocd import (
    BOCD,
    BOCDConfig,
)
from .cusum import (
    CUSUM,
    CUSUMConfig,
)
from .geometric_moving_average import (
    GeometricMovingAverage,
    GeometricMovingAverageConfig,
)
from .page_hinkley import (
    PageHinkley,
    PageHinkleyConfig,
)

__all__ = [
    "BOCD",
    "BOCDConfig",
    "CUSUM",
    "CUSUMConfig",
    "GeometricMovingAverage",
    "GeometricMovingAverageConfig",
    "PageHinkley",
    "PageHinkleyConfig",
]



"""Test concept drift detectors."""

from typing import Callable, Tuple

import numpy as np  # type: ignore
import pytest  # type: ignore

from frouros.detectors.concept_drift import ADWIN, ADWINConfig, KSWIN, KSWINConfig
from frouros.detectors.concept_drift import (
    BOCD,
    BOCDConfig,
    CUSUM,
    CUSUMConfig,
    GeometricMovingAverage,
    GeometricMovingAverageConfig,
    PageHinkley,
    PageHinkleyConfig,
)
from frouros.detectors.concept_drift import (
    DDM,
    DDMConfig,
    ECDDWT,
    ECDDWTConfig,
    EDDM,
    EDDMConfig,
    HDDMA,
    HDDMAConfig,
    HDDMW,
    HDDMWConfig,
    RDDM,
    RDDMConfig,
    STEPD,
    STEPDConfig,
)
from frouros.detectors.concept_drift.base import BaseConceptDrift
from frouros.detectors.concept_drift.streaming.change_detection.base import (
    BaseChangeDetection,
)
from frouros.detectors.concept_drift.streaming.change_detection.bocd import (
    GaussianUnknownMean,
)

MIN_NUM_INSTANCES = 30
BOCD_ARGS = {
    "model": GaussianUnknownMean(
        prior_mean=0,
        prior_var=1,
        data_var=0.5,
    ),
    "hazard": 0.01,
}
CUMSUM_ARGS = {
    "delta": 0.005,
    "lambda_": 50,
}
GEOMETRIC_MOVING_AVERAGE_ARGS = {
    "alpha": 0.99,
}
PAGE_HINKLEY_ARGS = {
    "alpha": 0.9999,
}
HDDM_ARGS = {
    "alpha_w": 0.005,
    "alpha_d": 0.001,
}
HDDMW_ARGS = {
    **HDDM_ARGS,
    "lambda_": 0.05,
}


def error_scorer(y_true, y_pred):
    """Error scorer function."""
    return int(1 - y_true == y_pred)


detectors = [
    (
        BOCD(
            config=BOCDConfig(
                min_num_instances=MIN_NUM_INSTANCES,
                **BOCD_ARGS,  # type: ignore
            ),
        ),
        error_scorer,
    ),
    (
        CUSUM(
            config=CUSUMConfig(
                min_num_instances=MIN_NUM_INSTANCES,
                **CUMSUM_ARGS,
            ),
        ),
        error_scorer,
    ),
    (
        GeometricMovingAverage(
            config=GeometricMovingAverageConfig(
                min_num_instances=MIN_NUM_INSTANCES,
                **GEOMETRIC_MOVING_AVERAGE_ARGS,
            ),
        ),
        error_scorer,
    ),
    (
        PageHinkley(
            config=PageHinkleyConfig(
                min_num_instances=MIN_NUM_INSTANCES,
                **PAGE_HINKLEY_ARGS,
            ),
        ),
        error_scorer,
    ),
    (
        ADWIN(
            config=ADWINConfig(
                clock=32,
                delta=0.15,
                m=5,
                min_num_instances=MIN_NUM_INSTANCES,
            ),
        ),
        error_scorer,
    ),
    (
        KSWIN(
            config=KSWINConfig(
                alpha=0.005,
                seed=31,
                min_num_instances=100,
                num_test_instances=30,
            ),
        ),
        error_scorer,
    ),
    (
        DDM(
            config=DDMConfig(
                warning_level=2.0,
                drift_level=3.0,
                min_num_instances=MIN_NUM_INSTANCES,
            ),
        ),
        error_scorer,
    ),
    (
        ECDDWT(
            config=ECDDWTConfig(
                lambda_=0.2,
                warning_level=0.5,
                min_num_instances=MIN_NUM_INSTANCES,
            ),
        ),
        error_scorer,
    ),
    (
        EDDM(
            config=EDDMConfig(
                alpha=0.95,
                beta=0.9,
                level=2.0,
                min_num_misclassified_instances=MIN_NUM_INSTANCES,
            ),
        ),
        error_scorer,
    ),
    (
        HDDMA(
            config=HDDMAConfig(
                **HDDM_ARGS,
                two_sided_test=False,
                min_num_instances=MIN_NUM_INSTANCES,
            ),
        ),
        error_scorer,
    ),
    (
        HDDMA(
            config=HDDMAConfig(
                **HDDM_ARGS,
                two_sided_test=True,
                min_num_instances=MIN_NUM_INSTANCES,
            ),
        ),
        error_scorer,
    ),
    (
        HDDMW(
            config=HDDMWConfig(
                **HDDM_ARGS,
                two_sided_test=False,
                min_num_instances=MIN_NUM_INSTANCES,
            ),
        ),
        error_scorer,
    ),
    (
        HDDMW(
            config=HDDMWConfig(
                **HDDM_ARGS,
                two_sided_test=True,
                min_num_instances=MIN_NUM_INSTANCES,
            ),
        ),
        error_scorer,
    ),
    (
        RDDM(
            config=RDDMConfig(
                warning_level=1.773,
                drift_level=2.258,
                max_concept_size=40000,
                min_concept_size=7000,
                max_num_instances_warning=1400,
                min_num_instances=129,
            ),
        ),
        error_scorer,
    ),
    (
        STEPD(
            config=STEPDConfig(
                alpha_d=0.003,
                alpha_w=0.05,
                min_num_instances=MIN_NUM_INSTANCES,
            ),
        ),
        lambda y_true, y_pred: y_true == y_pred,
    ),
]


@pytest.mark.parametrize("detector_info", detectors)
def test_streaming_detector_normal(
    clf_dataset: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
    train_prediction_normal: np.ndarray,
    detector_info: Tuple[BaseConceptDrift, Callable],
) -> None:
    """Test streaming detector.

    :param clf_dataset: dataset generated using SEA
    :type clf_dataset: Tuple[numpy.ndarray, numpy.ndarray,
    numpy.ndarray, numpy.ndarray]
    :param train_prediction_normal: test prediction values
    :type train_prediction_normal: numpy.ndarray
    :param detector_info: concept drift detector and value function
    :type detector_info: Tuple[BaseConceptDrift, Callable]
    """
    _, _, _, y_test = clf_dataset
    y_pred = train_prediction_normal  # noqa: N806
    detector, value_func = detector_info

    for y_sample_pred, y_sample in zip(y_pred, y_test):  # noqa: N806
        value_score = value_func(y_true=y_sample, y_pred=y_sample_pred)
        detector.update(value=value_score)


CHANGE_DETECTION_MIN_NUM_INSTANCES = 1
change_detection_detectors = [
    (
        BOCD(
            config=BOCDConfig(
                min_num_instances=CHANGE_DETECTION_MIN_NUM_INSTANCES,
                **BOCD_ARGS,  # type: ignore
            ),
        ),
        [100, 203],
    ),
    (
        CUSUM(
            config=CUSUMConfig(
                min_num_instances=CHANGE_DETECTION_MIN_NUM_INSTANCES,
                **CUMSUM_ARGS,
            ),
        ),
        [113, 226],
    ),
    (
        GeometricMovingAverage(
            config=GeometricMovingAverageConfig(
                min_num_instances=CHANGE_DETECTION_MIN_NUM_INSTANCES,
                **GEOMETRIC_MOVING_AVERAGE_ARGS,
            ),
        ),
        [134],
    ),
    (
        PageHinkley(
            config=PageHinkleyConfig(
                min_num_instances=CHANGE_DETECTION_MIN_NUM_INSTANCES,
                **PAGE_HINKLEY_ARGS,
            ),
        ),
        [113, 227],
    ),
]


@pytest.mark.parametrize("detector_info", change_detection_detectors)
def test_streaming_change_detection_detector(
    stream_drift: np.ndarray,
    detector_info: Tuple[BaseChangeDetection, list[int]],
) -> None:
    """Test streaming change detection detector.

    :param stream_drift: stream with drift
    :type stream_drift: numpy.ndarray
    :param detector_info: change detection detector and list of expected drift indices
    :type detector_info: Tuple[BaseChangeDetection, list[int]]
    """
    detector, idx_drifts = detector_info
    idx_detected_drifts = []
    for i, val in enumerate(stream_drift):
        detector.update(value=val)
        if detector.status["drift"]:
            detector.reset()
            idx_detected_drifts.append(i)

    assert idx_detected_drifts == idx_drifts



"""CVMTest (Cramér-von Mises test) module."""

from typing import Optional, Union

import numpy as np  # type: ignore
from scipy.stats import cramervonmises_2samp  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import NumericalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (
    BaseStatisticalTest,
    StatisticalResult,
)
from frouros.detectors.data_drift.exceptions import InsufficientSamplesError


class CVMTest(BaseStatisticalTest):
    """CVMTest (Cramér-von Mises test) [cramer1928composition]_ detector.

    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]

    :Note:
    - Passing additional arguments to `scipy.stats.cramervonmises_2samp <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.cramervonmises_2samp.html>`__ can be done using :func:`compare` kwargs.

    :References:

    .. [cramer1928composition] Cramér, Harald.
        "On the composition of elementary errors: First paper: Mathematical deductions."
        Scandinavian Actuarial Journal 1928.1 (1928): 13-74.

    :Example:

    >>> from frouros.detectors.data_drift import CVMTest
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = CVMTest()
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    StatisticalResult(statistic=5.331699999999998, p_value=1.7705426014202885e-10)
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        super().__init__(
            data_type=NumericalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    @BaseStatisticalTest.X_ref.setter  # type: ignore[attr-defined]
    def X_ref(self, value: Optional[np.ndarray]) -> None:  # noqa: N802
        """Reference data setter.

        :param value: value to be set
        :type value: Optional[numpy.ndarray]
        """
        if value is not None:
            self._check_sufficient_samples(X=value)
            self._X_ref = value
            # self._X_ref_ = check_array(value)  # noqa: N806
        else:
            self._X_ref = None  # noqa: N806

    def _specific_checks(self, X: np.ndarray) -> None:  # noqa: N803
        self._check_sufficient_samples(X=X)

    @staticmethod
    def _check_sufficient_samples(X: np.ndarray) -> None:  # noqa: N803
        if X.shape[0] < 2:
            raise InsufficientSamplesError("Number of samples must be at least 2.")

    @staticmethod
    def _statistical_test(
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
        **kwargs,
    ) -> StatisticalResult:
        test = cramervonmises_2samp(
            x=X_ref,
            y=X,
            **kwargs,
        )
        test = StatisticalResult(
            statistic=test.statistic,
            p_value=test.pvalue,
        )
        return test



"""IncrementalKSTest (IncrementalKolmogorov-Smirnov test) module."""

# FIXME: There seem to be a bug on the treap DS. Uncomment all  # pylint: disable=fixme
#  the commented code lines when that is solved.

# from copy import deepcopy
from typing import Optional, Union

import numpy as np  # type: ignore
from scipy.stats._stats_py import (  # type: ignore
    _compute_prob_outside_square,
    _compute_outer_prob_inside_method,
)
from scipy.stats.distributions import kstwo  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.data_drift.base import NumericalData, UnivariateData
from frouros.detectors.data_drift.streaming.statistical_test.base import (
    BaseStatisticalTest,
    StatisticalResult,
)
from frouros.utils.data_structures import CircularQueue

# from frouros.utils.data_structures import Treap

# Value used in scipy
MAX_AUTO_N = 10000


class IncrementalKSTest(BaseStatisticalTest):
    """IncrementalKSTest (Incremental Kolmogorov-Smirnov test) [dos2016fast]_ detector.

    :param window_size: window size value
    :type window_size: int
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :References:

    .. [dos2016fast] dos Reis, Denis Moreira, et al.
        "Fast unsupervised online drift detection using incremental kolmogorov-smirnov
        test."
        Proceedings of the 22nd ACM SIGKDD international conference on knowledge
        discovery and data mining. 2016.

    :Example:

    >>> from frouros.detectors.data_drift import IncrementalKSTest
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = IncrementalKSTest(window_size=10)
    >>> _ = detector.fit(X=X)
    >>> for sample in Y:
    ...     test, _ = detector.update(value=sample)
    ...     if test is not None:
    ...         print(test.statistic, test.p_value)
    """

    def __init__(  # noqa: D107
        self,
        window_size: int,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            data_type=NumericalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )
        self.window_size = window_size
        self.gcd = None
        # self.treap = None
        self.X_queue = CircularQueue(max_len=self.window_size)

    @property
    def window_size(self) -> int:
        """Window size property.

        :return: window size
        :rtype: int
        """
        return self._window_size

    @window_size.setter
    def window_size(self, value: int) -> None:
        """Window size setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("window_size value must be greater than 0.")
        self._window_size = value

    def _fit(self, X: np.ndarray) -> None:  # noqa: N803
        if max(X.shape[0], self.window_size) <= MAX_AUTO_N:
            self.gcd = np.gcd(X.shape[0], self.window_size)

        # r = X.shape[0] / self.window_size
        # self.treap = Treap(r=r)  # type: ignore
        # for obs in X:
        #     self.treap.insert(obs=obs, group=0)  # type: ignore
        self.X_ref = np.sort(X)

    def _reset(self) -> None:
        # self.treap = None
        self.gcd = None
        self.X_queue.clear()

    def _update(self, value: Union[int, float]) -> Optional[StatisticalResult]:
        self.X_queue.enqueue(value=value)
        # self.treap.insert(obs=value, group=1)  # type: ignore

        if self.num_instances < self.window_size:
            return None

        test = self._statistical_test(
            X_ref=self.X_ref,
            X=np.array(self.X_queue),
            gcd=self.gcd,
            window_size=self.window_size,
            # treap=self.treap,  # type: ignore
            # p_value=self._target_p_value
        )
        # self.treap.remove(obs=self.X_queue[0], group=1)  # type: ignore
        return test

    @staticmethod
    def _statistical_test(
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
        **kwargs,
    ) -> StatisticalResult:
        # treap, p_value = kwargs["treap"], kwargs["p_value"]

        # Resampling if |A| > |B|
        # samples_diff = max(X_ref.shape[0] - X.shape[0], 0)
        # if samples_diff > 0:
        #     treap_resample = deepcopy(treap)
        #     for _ in range(samples_diff):
        #         sample = np.random.choice(X, replace=True)
        #         treap_resample.insert(obs=sample, group=1)
        # statistic = max(
        #     treap_resample.max,
        #     -treap_resample.min
        # ) / treap_resample.num_samples[0]
        # else:
        #     statistic = max(treap.max, -treap.min) / treap.num_samples[0]
        # n, m = treap.num_samples
        # FIXME: Implement a CircularQueue that maintain   # pylint: disable=fixme
        #  values ordered
        X_sorted = np.sort(X)  # noqa: N806

        statistic = IncrementalKSTest._calculate_statistic(X_ref=X_ref, X=X_sorted)

        p_value = (
            IncrementalKSTest._calculate_p_value_exact(
                X_ref_num_samples=X_ref.shape[0],
                statistic=statistic,
                gcd=kwargs["gcd"],
                window_size=kwargs["window_size"],
            )
            if max(X_ref.shape[0], X_sorted.shape[0]) <= MAX_AUTO_N
            else IncrementalKSTest._calculate_p_value_aprox(
                X_ref_num_samples=X_ref.shape[0],
                X_num_samples=X.shape[0],
                statistic=statistic,
            )
        )

        test = StatisticalResult(statistic=statistic, p_value=p_value)
        return test

    @staticmethod
    def _calculate_p_value_aprox(
        X_ref_num_samples: int,  # noqa: N803
        X_num_samples: int,
        statistic: float,
    ) -> float:
        # Uses scipy code adaptation to calculate approximate p-value
        n, m = float(X_ref_num_samples), float(X_num_samples)
        en = n * m / (n + m)
        p_value = kstwo.sf(statistic, np.round(en))[0]
        return p_value

    @staticmethod
    def _calculate_statistic(X_ref: np.ndarray, X: np.ndarray) -> float:  # noqa: N803
        # Uses scipy code adaptation to calculate statistic (distance)
        data_all = np.concatenate([X_ref, X])
        cdf1 = np.searchsorted(X_ref, data_all, side="right") / X_ref.shape[0]
        cdf2 = np.searchsorted(X, data_all, side="right") / X.shape[0]
        cddiffs = cdf1 - cdf2
        argmin_s = np.argmin(cddiffs)
        argmax_s = np.argmax(cddiffs)
        min_s = np.clip(-cddiffs[argmin_s], 0, 1)
        max_s = cddiffs[argmax_s]
        statistic = min_s if min_s > max_s else max_s
        return statistic

    @staticmethod
    def _calculate_p_value_exact(
        X_ref_num_samples: int,  # noqa: N803
        statistic: float,
        gcd: Union[int, float],
        window_size: int,
    ) -> float:
        # Uses scipy code adaptation to calculate exact p-value
        lcm = (X_ref_num_samples // gcd) * window_size
        h = int(np.round(statistic * lcm))
        if h == 0:
            return 1.0
        try:
            with np.errstate(invalid="raise", over="raise"):
                p_value = (
                    _compute_prob_outside_square(X_ref_num_samples, h)
                    if X_ref_num_samples == window_size
                    else _compute_outer_prob_inside_method(
                        X_ref_num_samples, window_size, gcd, h
                    )
                )
        except (FloatingPointError, OverflowError):
            return np.nan
        return p_value



"""Base data drift module."""


import abc
import operator
from typing import Any, Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.base import BaseCallback
from frouros.detectors.base import BaseDetector
from frouros.detectors.data_drift.exceptions import DimensionError, MissingFitError


class BaseResult(abc.ABC):
    """Abstract class representing a result."""


class BaseDataType(abc.ABC):
    """Abstract class representing a data type."""

    @abc.abstractmethod
    def __init__(self) -> None:
        """Init method."""

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return (
            f"{self.__class__.__name__}"
            f"({', '.join(f'{k}={v}' for k, v in self.__dict__.items())})"
        )


class CategoricalData(BaseDataType):
    """Class representing categorical data."""

    def __init__(self) -> None:
        """Init method."""
        super().__init__()
        self.output_type = None


class NumericalData(BaseDataType):
    """Class representing numerical data."""

    def __init__(self) -> None:
        """Init method."""
        super().__init__()
        self.output_type = np.float32


class BaseStatisticalType(abc.ABC):
    """Abstract class representing a statistical data type."""

    @abc.abstractmethod
    def __init__(self) -> None:
        """Init method."""

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return (
            f"{self.__class__.__name__}"
            f"({', '.join(f'{k[1:]}={v}' for k, v in self.__dict__.items())})"
        )


class UnivariateData(BaseStatisticalType):
    """Class representing a univariate data type."""

    def __init__(self) -> None:
        """Init method."""
        super().__init__()
        self.dim_check = operator.eq


class MultivariateData(BaseStatisticalType):
    """Class representing a multivariate data type."""

    def __init__(self) -> None:
        """Init method."""
        super().__init__()
        self.dim_check = operator.ge


class BaseDataDrift(BaseDetector):
    """Abstract class representing a data drift detector."""

    def __init__(
        self,
        data_type: BaseDataType,
        statistical_type: BaseStatisticalType,
        callbacks: Optional[Union[BaseCallback, list[BaseCallback]]] = None,
    ) -> None:
        """Init method.

        :param data_type: data type
        :type data_type: BaseDataType
        :param statistical_type: statistical type
        :type statistical_type: BaseStatisticalType
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallback, list[Callback]]]
        """
        super().__init__(callbacks=callbacks)
        self.data_type = data_type
        self.statistical_type = statistical_type
        self.X_ref = None  # type: ignore

    @property
    def data_type(self) -> BaseDataType:
        """Data type property.

        :return: data type
        :rtype: BaseDataType
        """
        return self._data_type

    @data_type.setter
    def data_type(self, value: BaseDataType) -> None:
        """Data type setter.

        :param value: value to be set
        :type value: BaseDataType
        :raises TypeError: Type error exception
        """
        if not isinstance(value, BaseDataType):
            raise TypeError("value must be of type BaseDataType.")
        self._data_type = value

    @property
    def statistical_type(self) -> BaseStatisticalType:
        """Statistical type property.

        :return: statistical type
        :rtype: BaseStatisticalType
        """
        return self._statistical_type

    @statistical_type.setter
    def statistical_type(self, value: BaseStatisticalType) -> None:
        """Statistical type setter.

        :param value: value to be set
        :type value: BaseStatisticalType
        :raises TypeError: Type error exception
        """
        if not isinstance(value, BaseStatisticalType):
            raise TypeError("value must be of type BaseStatisticalType.")
        self._statistical_type = value

    @property
    def X_ref(self) -> Optional[np.ndarray]:  # noqa: N802
        """Reference data property.

        :return: reference data
        :rtype: Optional[numpy.ndarray]
        """
        return self._X_ref  # type: ignore # pylint: disable=E1101

    @X_ref.setter  # type: ignore
    def X_ref(self, value: Optional[np.ndarray]) -> None:  # noqa: N802
        """Reference data setter.

        :param value: value to be set
        :type value: Optional[numpy.ndarray]
        """
        if value is not None:
            self._check_array(X=value)
        self._X_ref = value

    def fit(self, X: np.ndarray, **kwargs) -> dict[str, Any]:  # noqa: N803
        """Fit detector.

        :param X: feature data
        :type X: numpy.ndarray
        :return: callbacks logs
        :rtype: dict[str, Any]
        """
        self._check_fit_dimensions(X=X)
        for callback in self.callbacks:  # type: ignore
            callback.on_fit_start(
                X=X,
            )
        self._fit(X=X, **kwargs)
        for callback in self.callbacks:  # type: ignore
            callback.on_fit_end(
                X=X,
            )

        callbacks_logs = self._get_callbacks_logs()
        return callbacks_logs

    def reset(self) -> None:
        """Reset method."""
        self.X_ref = None

    def _check_fit_dimensions(self, X: np.ndarray) -> None:  # noqa: N803
        try:
            if not self.statistical_type.dim_check(X.shape[1], 1):  # type: ignore
                raise DimensionError(f"Dimensions of X ({X.shape[-1]})")
        except IndexError as e:
            if not self.statistical_type.dim_check(X.ndim, 1):  # type: ignore
                raise DimensionError(f"Dimensions of X ({X.ndim})") from e

    def _check_is_fitted(self):
        if self.X_ref is None:
            raise MissingFitError("fit method has not been called")

    def _common_checks(self) -> None:  # noqa: N803
        self._check_is_fitted()

    @abc.abstractmethod
    def _fit(self, X: np.ndarray) -> None:  # noqa: N803
        pass

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return (
            f"{super().__repr__()[:-1]}, "
            f"data_type={self.data_type}, "
            f"statistical_type={self.statistical_type})"
        )



"""Batch callbacks init."""

from .permutation_test import PermutationTestDistanceBased
from .reset import ResetStatisticalTest

__all__ = [
    "PermutationTestDistanceBased",
    "ResetStatisticalTest",
]



"""Reset batch callback module."""

from typing import Any, Optional

import numpy as np  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.utils.logger import logger


class ResetStatisticalTest(BaseCallbackBatch):
    """Reset callback class that can be applied to :mod:`data_drift.batch.statistical_test <frouros.detectors.data_drift.batch.statistical_test>` detectors.

    :param alpha: significance value
    :type alpha: float
    :param name: name value, defaults to None. If None, the name will be set to `ResetStatisticalTest`.
    :type name: Optional[str]

    :Example:

    >>> from frouros.callbacks import ResetStatisticalTest
    >>> from frouros.detectors.data_drift import KSTest
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = KSTest(callbacks=ResetStatisticalTest(alpha=0.01))
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    INFO:frouros:Drift detected. Resetting detector...
    StatisticalResult(statistic=0.55, p_value=3.0406585087050305e-14)
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        alpha: float,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(name=name)
        self.alpha = alpha

    @property
    def alpha(self) -> float:
        """Alpha property.

        :return: significance value
        :rtype: float
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value: float) -> None:
        """Alpha setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value <= 0.0:
            raise ValueError("value must be greater than 0.")
        self._alpha = value

    def on_compare_end(
        self,
        result: Any,
        X_ref: np.ndarray,  # noqa: N803
        X_test: np.ndarray,
    ) -> None:
        """On compare end method.

        :param result: result obtained from the `compare` method
        :type result: Any
        :param X_ref: reference data
        :type X_ref: numpy.ndarray
        :param X_test: test data
        :type X_test: numpy.ndarray
        """
        p_value = result.p_value
        if p_value <= self.alpha:
            logger.info("Drift detected. Resetting detector...")
            self.detector.reset()  # type: ignore

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # def set_detector(self, detector) -> None:
    #     """Set detector method.
    #
    #     :raises TypeError: Type error exception
    #     """
    #     if not isinstance(detector, BaseDataDriftBatch):
    #         raise TypeError(
    #             f"callback {self.__class__.name} cannot be used with detector"
    #             f" {detector.__class__name}. Must be used with a detector of "
    #             f"type BaseDataDriftBatch."
    #         )
    #     self.detector = detector

    def reset(self) -> None:
        """Reset method."""



"""Base concept drift window based module."""

import abc
from typing import Union

from frouros.detectors.concept_drift.streaming.base import (
    BaseConceptDriftStreaming,
    BaseConceptDriftStreamingConfig,
)


class BaseWindowConfig(BaseConceptDriftStreamingConfig):
    """Class representing a window based configuration class."""


class BaseWindow(BaseConceptDriftStreaming):
    """Abstract class representing a window based."""

    config_type = BaseWindowConfig

    @abc.abstractmethod
    def _update(self, value: Union[int, float], **kwargs) -> None:
        pass



"""Bhattacharyya distance module."""

from typing import Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBasedBins,
)


class BhattacharyyaDistance(BaseDistanceBasedBins):
    """Bhattacharyya distance [bhattacharyya1946measure]_ detector.

    :param num_bins: number of bins in which to divide probabilities, defaults to 10
    :type num_bins: int
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallback, list[Callback]]]

    :References:

    .. [bhattacharyya1946measure] Bhattacharyya, Anil.
        "On a measure of divergence between two multinomial populations."
        Sankhyā: the indian journal of statistics (1946): 401-406.

    :Example:

    >>> from frouros.detectors.data_drift import BhattacharyyaDistance
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = BhattacharyyaDistance(num_bins=20)
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)
    DistanceResult(distance=0.2182101059622703)
    """

    def __init__(  # noqa: D107
        self,
        num_bins: int = 10,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        super().__init__(
            statistical_method=self._bhattacharyya,
            statistical_kwargs={
                "num_bins": num_bins,
            },
            callbacks=callbacks,
        )
        self.num_bins = num_bins

    def _distance_measure_bins(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
    ) -> float:
        bhattacharyya = self._bhattacharyya(X=X_ref, Y=X, num_bins=self.num_bins)
        return bhattacharyya

    @staticmethod
    def _bhattacharyya(
        X: np.ndarray, Y: np.ndarray, *, num_bins: int  # noqa: N803
    ) -> float:
        (  # noqa: N806
            X_percents,
            Y_percents,
        ) = BaseDistanceBasedBins._calculate_bins_values(
            X_ref=X, X=Y, num_bins=num_bins
        )
        bhattacharyya = 1 - np.sum(np.sqrt(X_percents * Y_percents))
        return bhattacharyya
